/** 
   @file JavaPlugin.h
 
    Implements MPlugInBuilder as defined in Plugin API
 
 	@author Tomaso Paoletti <tpaoletti@users.sourceforge.net>
 
 	Based on Translator sample plugin by John R. Dance, 1999
 	
   This software is distributed under the GNU Public License.  
   See section COPYING of the GNU Public license
   for conditions under which this file may be redistributed.
*/

#ifndef JAVA_PLUGIN_H
#define JAVA_PLUGIN_H

#include "MPlugInBuilder.h"
#include "ErrorMessage.h"

#include "JavaPluginPrefs.h"

#define DBG(expr)	printf##expr


class JavaPlugin : public MPlugInBuilder
{
public:
								JavaPlugin();
								~JavaPlugin();
									
	virtual status_t			GetToolName(
									MProject* inProject,
									char* 		outName,
									int32		inBufferLength,
									MakeStageT	inStage,
									MakeActionT	inAction);

	virtual const char *		LinkerName();
	virtual MakeStageT			MakeStages();
	virtual MakeActionT			Actions();
	virtual PlugInFlagsT		Flags();
	virtual ulong				MessageDataType();

	// return true if something changed in the settings
	virtual bool				ValidateSettings(
									BMessage&	inOutMessage);

	// return B_NO_ERROR if everything went OK
	virtual status_t			BuildPrecompileArgv(
									MProject& inProject,
									BList& 		inArgv,
									MFileRec& 	inFileRec);

	virtual status_t			BuildCompileArgv(
									MProject& inProject,
									BList& 		inArgv,
									MakeActionT inAction,
									MFileRec& 	inFileRec);

	virtual status_t			BuildPostLinkArgv(
									MProject& inProject,
									BList& 		inArgv,
									MFileRec& 	inFileRec);

	virtual bool				FileIsDirty(
									MProject& inProject,
									MFileRec& 	inFileRec,
									MakeStageT	inStage,
									MakeActionT	inAction,
									time_t		inModDate);

	virtual status_t			ParseMessageText(
									MProject& inProject,
									const char*	inText,
									BList&		outList);
	virtual void				CodeDataSize(
									MProject& inProject,
									const char* inFilePath,
									int32&		outCodeSize,
									int32&		outDataSize);
	virtual status_t			GenerateDependencies(
									MProject& inProject,
									const char* inFilePath,
									BList&		outList);

	virtual void				GetTargetFilePaths(
									MProject& inProject,
									MFileRec& 	inFileRec,
									BList&		inOutTargetFileList);

	virtual void				ProjectChanged(
									MProject& inProject,
									ChangeT		inChange);

private:
	//MProject		*fProject;
	compilerPrefs	fPrefs;
	
	BList *			fFiles;
	
	char *						getCompletePath(char * file);
				
	void						makeOutputFileName(
									const char *filePath, 
									BString &outputName);
									
	ErrorMessage*				parseFileLineError(const BString& text,
									const char* filePrefix, 
									const char* fileSuffix, 
									const char* linePrefix, 
									const char* lineSuffix);
									
	ErrorMessage*				createEmptyError();
	ErrorMessage*				createTextOnlyMessage(const BString& text);

	int							currentCompiler();
};

#endif /* JAVA_PLUGIN_H */
