/**
 * @file btn.cpp
 * Native buttons (push button, radio and check). 
 *
 * Copyright (c) 1998-2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * @author T. Paoletti <tpaoletti@users.sourceforge.net>,
 *         <your name here>
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "toolkit.hpp"

#include <Button.h>

#define PUSHBUTTON		1
#define CHECKBOX		2
#define RADIOBUTTON		4

extern char * RootViewName;

#define EVENT_CONTROL	'KCmd'


/** Native widget to implement push buttons, radio buttons and check boxen. */

class KButton:public KWnd {
public:	
	bool	blockCmd;
	int		fStyle;
	
	KButton( BView * parent, char* lbl, int style)
	{
/*		style |= DEF_WND_STYLE;
		hwnd = CreateWindow( TEXT("BUTTON"), lbl, style, 0, 0, 0, 0,
			parent, (HMENU)0, NULL, NULL);
		defWndProc = (WNDPROC)SetWindowLong( hwnd, GWL_WNDPROC, (LONG)WndProc);
		blockCmd = false;
*/
		BControl * v;
		BRect r(0,0,20,20);
		fStyle = style;
		switch(fStyle) {
			default: case PUSHBUTTON:
				v = new BButton(r, "KButton,push", lbl, new BMessage(EVENT_CONTROL));
				break;
			case CHECKBOX:
				v = new BCheckBox(r, "KButton,check", lbl, new BMessage(EVENT_CONTROL));
				break;
			case RADIOBUTTON:
				v = new BRadioButton(r, "KButton,radio", lbl, new BMessage(EVENT_CONTROL));
				break;
		}
				
		setView(v);	
		parent->AddChild(v);
		v->SetTarget(v);
		DBG(AWT, ("new KButton %p (parent %p): %s\n", v, parent, lbl)); 
	}

	/** Access to widget's view.*/
	BControl * NATIVE() {
		return (BControl *)getView();
	}
	
	bool KButton::OnCommand( int cmd) {
		DBG(AWT, ("KButton::OnCommand\n"));

		jstring aCmd;
	
		//Grouped CheckBox
		if ( fStyle & RADIOBUTTON ) {
			// workaround for notifications caused by WM_SETFOCUS

#if 0 /** @todo tab stepping is disabled */
			// ( tab stepping )
			if ( blockCmd ) {
				blockCmd = false;
				return false;
			}
#endif
/*
			X->jEvt = JniEnv->CallStaticObjectMethod( ItemEvent, getItemEvent,
				srcIdx, getCheck() ? DESELECTED : SELECTED, 0 );
*/
			X->jEvt = JniEnv->CallStaticObjectMethod( ItemEvent, getItemEvent,
				srcIdx, getCheck() ? SELECTED : DESELECTED, 0 );
		}
		//Simple CheckBox
		else if ( fStyle & CHECKBOX ) {
/*			X->jEvt = JniEnv->CallStaticObjectMethod( ItemEvent, getItemEvent,
				srcIdx, getCheck() ? DESELECTED : SELECTED, 0 );
*/
			X->jEvt = JniEnv->CallStaticObjectMethod( ItemEvent, getItemEvent,
				srcIdx, getCheck() ? SELECTED : DESELECTED, 0 );
		}
		//PushButton
		else {
			strncpy(X->buf, NATIVE()->Label(), X->nBuf);
			aCmd = JniEnv->NewStringUTF( CString2UTF(X, X->buf, strlen(X->buf)) );
			X->jEvt = JniEnv->CallStaticObjectMethod( ActionEvent, getActionEvent,
				srcIdx, aCmd, 0 );
		}

		return true;
	}
	
	bool KButton::OnFocusChange ( bool gain) {
		X->jEvt = JniEnv->CallStaticObjectMethod( FocusEvent, getFocusEvent,
			srcIdx, gain ? FOCUS_GAINED : FOCUS_LOST, false);
		// workaround for notifications caused by WM_SETFOCUS ( tab stepping )

UNIMPLEMENTED((
		blockCmd = true;
));
		return false;
	}
	

	bool KButton::OnMouseUp( int btn, int x, int y) {
		KWnd::OnMouseUp( btn, x, y);

		// workaround for notifications caused by WM_SETFOCUS ( tab stepping )
		blockCmd = false;
		return false;

	}

	void KButton::setEnabled( jboolean en) {
		DBG(AWT, ("btnSetEnabled: %p %d\n", this, en));
		if (lock()) {
			NATIVE()->SetEnabled( en);
			unlock();
		}
	}

	void KButton::setCheck( jboolean check) {
		if (lock()) {
			NATIVE()->SetValue( check ? B_CONTROL_ON : B_CONTROL_OFF);
			unlock();
		}
	}
	
	bool KButton::getCheck() {
		bool check = false;
		if (lock()) {
			check = NATIVE()->Value() != B_CONTROL_OFF;
			unlock();
		}
		return check;
	}
	
	void KButton::setText(jstring text) {
		if (lock()) {
			if ( text) {
				NATIVE()->SetLabel( java2CString( JniEnv, X, text));
			} else {
				NATIVE()->SetLabel( "");
			}
			unlock();
		}
	}
	
	void KButton::setGrouped( jboolean grouped) {
UNIMPLEMENTED((
		int style = GetWindowLong( hwnd, GWL_STYLE);
		if ( grouped) {
			style |= BS_RADIOBUTTON;
			style &= ~BS_CHECKBOX;
		}
		else {
			style |= BS_CHECKBOX;
			style &= ~BS_RADIOBUTTON;
		}
		SetWindowLong( hwnd, GWL_STYLE, style);
));
	}
};



/*************************
*	exported functions
**************************/
extern "C" {
	
	void* __cdecl
		Java_java_awt_Toolkit_btnCreatePushbutton ( JNIEnv* env, jclass clazz, KWnd* parent, jstring label )
	{
		return ( new KButton( parent->getView(), java2CString( env, X, label), PUSHBUTTON) );
	}
	
	void* __cdecl
		Java_java_awt_Toolkit_btnCreateCheckbox ( JNIEnv* env, jclass clazz, KWnd* parent, jstring label )
	{
		return ( new KButton( parent->getView(), java2CString( env, X, label), CHECKBOX) );
	}
	
	void* __cdecl
		Java_java_awt_Toolkit_btnCreateRadiobutton ( JNIEnv* env, jclass clazz, KWnd* parent, jstring label )
	{
		return ( new KButton( parent->getView(), java2CString( env, X, label), RADIOBUTTON) );
	}
	
	void __cdecl
		Java_java_awt_Toolkit_btnSetLabel ( JNIEnv* env, jclass clazz, KButton* wnd, jstring label)
	{
		wnd->setText( label);
	}
	
	void __cdecl
		Java_java_awt_Toolkit_btnSetCheck ( JNIEnv* env, jclass clazz, KButton* wnd, jboolean check )
	{
		DBG(AWT, ("btnSetCheck %p %d\n", wnd, check));
		wnd->setCheck( check);
	}
	
	jboolean __cdecl
		Java_java_awt_Toolkit_btnGetCheck ( JNIEnv* env, jclass clazz, KButton* wnd )
	{
		DBG(AWT, ("btnGetCheck %p\n", wnd));
		return wnd->getCheck();
	}
	
	void __cdecl
		Java_java_awt_Toolkit_btnSetGrouped ( JNIEnv* env, jclass clazz, KButton* wnd, jboolean grouped )
	{
		wnd->setGrouped( grouped);
	}
}
