/**
 * @file choice.cpp 
 * Native multiple choice object. 
 *	
 * Copyright (c) 1998-2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * @author T. Paoletti <tpaoletti@users.sourceforge.net>,
 *         <your name here>
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */
#include "toolkit.hpp"


/** Multiple choice widget.

    @todo implement ItemEvent in response to mouse selection
    @todo review KChoice::getPreferredSize()
    @todo review KChoice::setBounds()
*/
class KChoice:public KWnd {

BPopUpMenu *	fItems;

public:		
	KChoice( BView * parent) {

		BRect r(0,0,0,0);
		
		fItems = new BPopUpMenu("");	
		fItems->SetLabelFromMarked(true);
		BView * v = new BMenuField(r, "KChoice", "TEXT", fItems );
		
		setView(v);
		parent->AddChild(v);
	}
	
	BMenuField * NATIVE() {
		return (BMenuField *)getView();
	}
	
	bool KChoice::OnCommand( int cmd) {
UNIMPLEMENTED((
		int idx;
		
		switch ( cmd) {
		case CBN_CLOSEUP:
			break;
		case CBN_DROPDOWN:
			break;
		case CBN_EDITCHANGE:
			break;
		case CBN_EDITUPDATE:
			break;
		case CBN_SELCHANGE:
			break;
		case CBN_SELENDCANCEL:
			break;
		case CBN_SELENDOK:
			idx = getSelectionIdx();
			X->jEvt = JniEnv->CallStaticObjectMethod( ItemEvent, getItemEvent,
				srcIdx, SELECTED, idx );
			break;
		case CBN_DBLCLK:
			break;
		case CBN_SETFOCUS:
			break;
		case CBN_KILLFOCUS:
			break;
		case CBN_ERRSPACE:
			break;
		}
));		
		return true;
	}
	
	jobject KChoice::getPreferredSize() {
		jclass dc = JniEnv->FindClass( "java/awt/Dimension");
		jmethodID ct = JniEnv->GetMethodID( dc, "<init>", "(II)V;");
UNIMPLEMENTED((
		int height = SendMessage( hwnd, CB_GETITEMHEIGHT, (WPARAM)-1, 0);
		return (JniEnv->NewObject( dc, ct, 100, height));
));
		return (JniEnv->NewObject( dc, ct, 100, 100));
	}
	
	void KChoice::setBounds(int x, int y, int width, int height) {
UNIMPLEMENTED((
		//extends include pull down list
		int ih = (int)SendMessage( hwnd, CB_GETITEMHEIGHT, 0, 0);
		KWnd::setBounds( x, y, width, height + 8*ih);
		SendMessage( hwnd, CB_SETITEMHEIGHT, (WPARAM)-1, (LPARAM)height);
));
		KWnd::setBounds( x, y, width, height /* !!! + 8*ih */);
	}
	
	int KChoice::getSelectionIdx() {		
		BMenuItem * item = fItems->FindMarked();
		
		return (item!=NULL ? fItems->IndexOf(item) : 0);
	}

	void KChoice::addItem(char * txt) { 	
		if (lock()) {
			fItems->AddItem(new BMenuItem(txt, NULL));
			unlock();
		}		
	}
	
	void KChoice::addItem(char * txt, int32 index) { 	
		if (lock()) {
			fItems->AddItem(new BMenuItem(txt, NULL), index);
			unlock();
		}		
	}
	
	void KChoice::selectItem(int32 index) {
		if (lock()) {
			BMenuItem * mi = fItems->ItemAt(index);
			if (mi)
				mi->SetMarked(true);
			unlock();
		}	
	}
	
	void KChoice::removeItem(int32 index) { 	
		if (lock()) {
			fItems->RemoveItem(index);
			unlock();
		}		
	}

	void KChoice::removeAll() { 	
		if (lock()) {
			fItems->RemoveItems(0, fItems->CountItems(), true);
			unlock();
		}		
	}
};

/*************************
*	exported functions
**************************/
extern "C" {
	
	void* __cdecl
		Java_java_awt_Toolkit_choiceCreateChoice ( JNIEnv* env, jclass clazz, KWnd* parent)
	{
		return ( new KChoice( parent->getView()));
	}
	
	void __cdecl
		Java_java_awt_Toolkit_choiceAppendItem ( JNIEnv* env, jclass clazz, KChoice* wnd, jstring item)
	{
		char* buf = java2CString( env, X, item);
		wnd->addItem(buf); 	
	}
	
	void __cdecl
		Java_java_awt_Toolkit_choiceInsertItem ( JNIEnv* env, jclass clazz, KChoice* wnd, jstring item, jint idx)
	{
		char* buf = java2CString( env, X, item);
		wnd->addItem(buf, idx); 	
	}
	
	void __cdecl
		Java_java_awt_Toolkit_choiceSelectItem ( JNIEnv* env, jclass clazz, KChoice* wnd, jint idx)
	{
		wnd->selectItem(idx);
	}
	
	void __cdecl
		Java_java_awt_Toolkit_choiceRemoveAll ( JNIEnv* env, jclass clazz, KChoice* wnd)
	{
		wnd->removeAll();
	}
	
	void __cdecl
		Java_java_awt_Toolkit_choiceRemoveItem ( JNIEnv* env, jclass clazz, KChoice* wnd, jint idx)
	{
		wnd->removeItem(idx);
	}
	
	
}
