/**
 * @file cmnwnd.cpp
 * Basic view functions
 *
 * Copyright (c) 1999-2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * @author T. Paoletti <tpaoletti@users.sourceforge.net>,
 *         <your name here>
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "toolkit.hpp"


extern BMessage * KWM_REQUEST_FOCUS;
extern BMessage * KWM_DESTROY;


	/** Locks the native component's looper.
	    Sports a timeout of 5 secs and some debugging aids */ 
bool KWnd::lock() {
	bool res = fView->LockLooperWithTimeout(5e6) == B_OK;		/* !!! CHECK VALUE */
	if (!res) printf("LOCK TIMED OUT on %p (owner is %d)\n", getView(), getWindow()->LockingThread() );
	return res;
}

void KWnd::unlock() {
	fView->UnlockLooper( );
}


bool KWnd::OnCommand( int cmd) {
	DBG(AWT, ("KWnd::OnCommand does nothing\n"));
//	if ( popup )
//		processMenuAction( popup, cmd);
	return false;
}

bool KWnd::OnScroll( int pos, int op, bool vert) {
	return false;
}

bool KWnd::OnSetCursor() {
// !!!	be_app->SetCursor( cursor);
	return true;
}

bool KWnd::OnActivate( int flags, bool min, BWindow * handle ) {
	return false;
}

bool KWnd::OnPosChanged( int x, int y, int width, int height) {
	// done by java layouting
	return false;
}

bool KWnd::OnPaint( int x, int y, int width, int height){

	/*
	* paint native *before* java painting
	*/
/** @todo review dispatchNativeEvent */
//		dispatchNativeEvent();
	
// 		CallWindowProc( defWndProc, hwnd, WM_PAINT, 0, 0);

	X->jEvt = JniEnv->CallStaticObjectMethod( PaintEvent, getPaintEvent,
		srcIdx, PAINT,
		x, y, width, height );			
	
	return true;
}

bool KWnd::OnOwnerDraw( /* LPDRAWITEMSTRUCT di */) {
	return false;
}

bool KWnd::OnEraseBkgnd( BView * view){
	return false;
}

HBRUSH KWnd::OnCtlColor( BView * view) {
UNIMPLEMENTED((
	SetBkColor( hdc, bg);
	SetTextColor( hdc, fg);
	return brush;
));
	return NULL;
}

bool KWnd::OnContextMenu( int x, int y) {
UNIMPLEMENTED((
	POINT pt = { x, y };
	int retID;
	
	if ( popup ) {
		MapWindowPoints( hwnd, HWND_DESKTOP, &pt, 1);
		retID = TrackPopupMenuEx( popup, TPM_LEFTALIGN | TPM_TOPALIGN | TPM_RETURNCMD,
			pt.x, pt.y, hwnd, NULL);
		processMenuAction( popup, retID);
	}
));	
	return true;
}

bool KWnd::OnMouseDown( int btn, int x, int y){
	int jBtn = ( btn & B_SECONDARY_MOUSE_BUTTON ) ? 3 : 1;

	X->jEvt = JniEnv->CallStaticObjectMethod( MouseEvent, getMouseEvent,
		srcIdx, MOUSE_PRESSED,
		jBtn, x+(int)insets.left, y+(int)insets.top );
	return false;
}

bool KWnd::OnMouseUp( int btn, int x, int y){
	int jBtn = ( btn & B_SECONDARY_MOUSE_BUTTON ) ? 3 : 1;

	X->jEvt = JniEnv->CallStaticObjectMethod( MouseEvent, getMouseEvent,
		srcIdx, MOUSE_RELEASED,
		jBtn, x+(int)insets.left, y+(int)insets.top );
	return false;
}

bool KWnd::OnMouseEnter( int x, int y){
	return false;
}

bool KWnd::OnMouseExit( int x, int y){
	return false;
}

bool KWnd::OnMouseMove( int x, int y){
	DBG(AWT, ("OnMouseMove %d %d\n", x,y));

	X->jEvt = JniEnv->CallStaticObjectMethod( MouseEvent, getMouseEvent,
		srcIdx, J_MOUSE_MOVED,
		1, x+(int)insets.left, y+(int)insets.top);
	return false;
}

bool KWnd::OnFocusChange ( bool gain) {
	X->jEvt = JniEnv->CallStaticObjectMethod( FocusEvent, getFocusEvent,
		srcIdx, gain ? FOCUS_GAINED : FOCUS_LOST, false);
	return false;
}

bool KWnd::OnChar( TCHAR c, int data) {
	return false;
}

bool KWnd::OnKeyDown( int virtKey, int data) {
	int scan, chr, mods;

UNIMPLEMENTED((
	mods = convertKey( virtKey, &scan, &chr);
));
	scan = virtKey;
	chr = virtKey;
	mods = 0;
	//kprintf( "OnKeyDown: scan:%d[%x] data:%x chr:%d[%x] mods:%x\n", scan, scan, (data & 0x00ff0000) >> 16, chr, chr, mods);

	X->jEvt = JniEnv->CallStaticObjectMethod( KeyEvent, getKeyEvent,
		srcIdx, KEY_PRESSED, scan, chr, mods );
	return false;
}

bool KWnd::OnKeyUp( int virtKey, int data) {
	int scan, chr, mods;
	
UNIMPLEMENTED((
	mods = convertKey( virtKey, &scan, &chr);
));
	scan = virtKey;
	chr = virtKey;
	mods = 0;
	//kprintf( "OnKeyUp: scan:%d[%x] chr:%d[%x] mods:%x\n", scan, scan, chr, chr, mods);
	
	X->jEvt = JniEnv->CallStaticObjectMethod( KeyEvent, getKeyEvent,
		srcIdx, KEY_RELEASED, scan, chr, mods );
	return false;
}

bool KWnd::OnClose() {
	return false;
}

int KWnd::getKeyMods() {
	int mods = 0;
	int32 nmods = modifiers();
	if (nmods & B_SHIFT_KEY)	mods |= 1;
	if (nmods & B_CONTROL_KEY)	mods |= 2;
	if (nmods & B_MENU_KEY)		mods |= 8;
	return mods;
}

int KWnd::convertKey( int virtKey, int* scan, int* chr ) {
UNIMPLEMENTED((
	static BYTE states[256];
	int mods, cpy;
	WORD ac;

	mods = getKeyMods();
	GetKeyboardState( states);
	cpy = ToAscii( virtKey, 0, states, &ac, 0);

	if ( cpy == 0 ) {
		*chr = 0;

		switch ( virtKey ) {
		case VK_DELETE:
			virtKey = 0x7F;
			break;
		case VK_INSERT:
			virtKey = 0x9B;
			break;
		case VK_TAB:
			//ctrl-tab KEY_TYPED generation ( tabbing in MLE )
			*chr = '\t';
			break;
		}
		*scan = virtKey;
	}
	else if ( cpy < 0 ) {	//dead key
	}
	else {
		ac &= 0x00FF;
		switch ( ac) {
		case VK_RETURN:
			virtKey = ac = '\n';
			break;
		}
		*scan = virtKey;
		*chr = ac;
	}

	return mods;

	/*
	if ( *chr = MapVirtualKey( virtKey, 2) ) {
		switch ( *chr ) {
		case VK_RETURN:
			*chr = '\n';
			break;
		}

		*scan = *chr;
		if ( (mods & 1) == 0 ) {
			if ( IsCharUpper( *chr) )
				*chr |= 0x20;
		}
	}
	else {
		switch ( virtKey ) {
		case VK_DELETE:
			virtKey = 0x7F;
			break;
		case VK_INSERT:
			virtKey = 0x9B;
			break;
		}
		*scan = virtKey;
		*chr = 0;
	}
	return mods;
	*/
));
}

jobject KWnd::getPreferredSize() {
	jclass dc = JniEnv->FindClass( "java/awt/Dimension");
	jmethodID ct = JniEnv->GetMethodID( dc, "<init>", "(II)V;");
	return (JniEnv->NewObject( dc, ct, 10, 10));
}

void KWnd::invalidate() {
	getView()->Invalidate();
}

/** There is no generic BView method to set the object text,
    ONLY SUBCLASSES OF KWnd IMPLEMENT THIS
*/
void KWnd::setText( jstring text) {
DBG(AWT, ("KWnd::setText does nothing\n"));
/*
	if ( text) {
		SetWindowText( hwnd, java2CString( JniEnv, X, text));
	}
	else {
		SetWindowText( hwnd, TEXT("") );
	}
*/
}

/** There is no generic BView method to get the object text,
    ONLY SUBCLASSES OF KWnd IMPLEMENT THIS
*/
char* KWnd::getText() {
DBG(AWT, ("KWnd::getText does nothing\n"));
/* SUBCLASSES WILL IMPLEMENT THIS
	UINT len = GetWindowTextLength( hwnd);
	getBuffer(X, len+1);
	GetWindowText( hwnd, X->buf, len+1);
	return X->buf;
*/
	return NULL;
}

/** ONLY SUBCLASSES OF KWnd IMPLEMENT THIS */
void KWnd::setEnabled( jboolean b) {
DBG(AWT, ("KWnd::setEnabled does nothing\n"));
}

void KWnd::setVisible( jboolean b) {
	if (lock()) {
		if (b) {
			fView->Show(); 
		} else {
			fView->Hide();
		}
		unlock();
	}
}

void KWnd::setBounds(int x, int y, int width, int height) {
	KWnd* parent = NULL; // !!! = (KWnd*)GetWindowLong( GetParent( hwnd), GWL_USERDATA );

	DBG(AWT_GRA, ("setBounds: %p %d,%d,%d,%d\n", this, x,y,x+width,y+height));
	
	if (lock()) {		
		if ( parent ) {
			x -= (int)parent->insets.left;
			y -= (int)parent->insets.top;
		}
			
		fView->MoveTo( x, y);
		fView->ResizeTo( width, height); 
		finishResize();		// finish resize of widget
		unlock();
	}
}

void KWnd::finishResize() {
	DBG(AWT_GRA, ("finishResize\n"));
}

void KWnd::setForeground(rgb_color c) {
	DBG(AWT_GRA, ("setForeground: %p %x\n", this, c));
	fg = c;
	/** @todo OnCtlColor IMPLEMENTATION */
	getView()->SetHighColor(c);
}

void KWnd::setBackground(rgb_color c) {
	DBG(AWT_GRA, ("setBackground: %p %x\n", this, c));
	bg = c;
	getView()->SetViewColor(c);
}

void KWnd::setFont(BFont * f) {
	fnt = f;
	getView()->SetFont(f);
}

void KWnd::setCursor(BCursor * cursor) {
// !!!	this->cursor = cursor;
}

void KWnd::setFocus() {
UNIMPLEMENTED((
	postEvent( KWM_REQUEST_FOCUS);
));
	getView()->MakeFocus(true);
}

void KWnd::postEvent(BMessage * msg)
{
	getWindow()->PostMessage(msg);
}
 
bool KWnd::processMenuAction( int id) {
UNIMPLEMENTED((
	MENUITEMINFO mi;
	jmethodID getEvt;
	
	if ( menu ) {
		mi.cbSize = sizeof( MENUITEMINFO);
		mi.fMask = MIIM_DATA;
		
		if ( GetMenuItemInfo( menu, id, false, &mi) ) {
			getEvt = JniEnv->GetStaticMethodID( ActionEvent, "getMenuEvent",
				"(Ljava/awt/MenuItem;)Ljava/awt/ActionEvt;");
			JniEnv->CallStaticObjectMethod( ActionEvent, getEvt, (jobject)mi.dwItemData);
			
			return true;
		}
	}
));
	jmethodID getEvt;
	
	//if ( menu ) {
	DBG(AWT, ("processMenuAction: %p %d\n", this, id));		
		getEvt = JniEnv->GetStaticMethodID( ActionEvent, "getMenuEvent",
			"(Ljava/awt/MenuItem;)Ljava/awt/ActionEvt;");
		X->jEvt = JniEnv->CallStaticObjectMethod( ActionEvent, getEvt, id);
			
		return true;
	//}

	//return false;
}

bool KWnd::updateInsets() {
	return false;
}


/*************************
*	exported functions
**************************/
extern "C" {
	
	void __cdecl
		Java_java_awt_Toolkit_cmnDestroyWindow ( JNIEnv* env, jclass clazz, KWnd* wnd )
	{
UNIMPLEMENTED((
		wnd->postEvent(KWM_DESTROY);
));
	}

	void* __cdecl
		Java_java_awt_Toolkit_cmnGetPreferredSize( JNIEnv* env, jclass clazz, KWnd* wnd)
	{
		return wnd->getPreferredSize();
	}
	
	void __cdecl
		Java_java_awt_Toolkit_cmnSetBounds ( JNIEnv* env, jclass clazz, KWnd* wnd, jint x, jint y,
		jint width, jint height)
	{
		DBG(AWT_GRA, ("cmnSetBounds: %p %d,%d,%d,%d\n", wnd, x,y,x+width,y+height));
		
		if (wnd->lock()) {
			wnd->setBounds( x, y, width, height);
			wnd->invalidate();
			wnd->unlock();
		}
	}
	
	void __cdecl
		Java_java_awt_Toolkit_cmnSetForeground ( JNIEnv* env, jclass clazz, KWnd* wnd, jint clr)
	{
		DBG(AWT_GRA, ("cmnSetForeground: %p %x\n", wnd, clr));
		if (wnd->lock()) {
			wnd->setForeground( RGBCAST(clr));	
			wnd->invalidate();
			wnd->unlock();
		}
	}
	
	void __cdecl
		Java_java_awt_Toolkit_cmnSetBackground ( JNIEnv* env, jclass clazz, KWnd* wnd, jint clr)
	{
		DBG(AWT_GRA, ("cmnSetBackground: %p %x\n", wnd, clr));
		if (wnd->lock()) {
			wnd->setBackground( RGBCAST(clr));	
			wnd->invalidate();
			wnd->unlock();
		}
	}
	
	void __cdecl
		Java_java_awt_Toolkit_cmnSetFont ( JNIEnv* env, jclass clazz, KWnd* wnd, Font* fnt)
	{
		DBG(AWT_GRA, ("cmnSetFont: %p %p\n", wnd, fnt));
		if (wnd->lock()) {
			wnd->setFont( fnt);
			wnd->invalidate();
			wnd->unlock();
		}
	}
	
	void __cdecl
		Java_java_awt_Toolkit_cmnSetCursor ( JNIEnv* env, jclass clazz, KWnd* wnd, jint cursor)
	{
		wnd->setCursor( getCursor( cursor));	
	}
	
	void __cdecl
		Java_java_awt_Toolkit_cmnSetFocus ( JNIEnv* env, jclass clazz, KWnd* wnd)
	{
		DBG(AWT_GRA, ("cmnSetFocus: %p\n", wnd));
		if (wnd->lock()) {
			wnd->setFocus();
			wnd->unlock();
		}	
	}
	
	void __cdecl
		Java_java_awt_Toolkit_cmnSetVisible ( JNIEnv* env, jclass clazz, KWnd* wnd, jboolean vis)
	{
		DBG(AWT_GRA, ("cmnSetVisible: %p %d\n", wnd, vis));
		wnd->setVisible( vis);	/* NO EXPLICIT LOCK HERE, OVERLOADED VERSIONS WILL LOCK AS NEEDED */
	}
	
	void __cdecl
		Java_java_awt_Toolkit_cmnSetEnabled ( JNIEnv* env, jclass clazz, KWnd* wnd, jboolean en)
	{
		DBG(AWT_GRA, ("cmnSetEnabled: %p %d\n", wnd, en));
		if (wnd->lock()) {
			wnd->setEnabled( en);
			wnd->unlock();
		}
	}
	
	void __cdecl
		Java_java_awt_Toolkit_cmnValidate( JNIEnv* env, jclass clazz, KWnd* wnd)
	{
		DBG(AWT, ("cmnValidate %p\n", wnd));
UNIMPLEMENTED((
		UpdateWindow( wnd->hwnd);
));
		if (wnd->lock()) {
			wnd->invalidate();
			wnd->unlock();
		}
	}
	
}
