/**
 * @file edit.cpp 
 * Text edit classes
 *
 * Copyright (c) 1998-2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * @author T. Paoletti <tpaoletti@users.sourceforge.net>,
 *         <your name here>
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */
 
#include "toolkit.hpp"
#include <TextView.h>

#define VK_TAB		'\t'
#define VK_RETURN	'\n'

/*
#define ESB_BOTH 0
#define ESB_VERT 1
#define ESB_HORZ 2
#define ESB_NONE 3
*/

/** 
	Native widget for text fields and areas 
*/
class KEdit:public KWnd {
public:		
	bool multi;

	KEdit( BView * parent, char* txt, int style) {

		BRect r(0,0,0,0);
		BTextView * v = new BTextView(r, "KEdit", r, B_FOLLOW_ALL_SIDES );
		parent->AddChild(v);
		setView(v);

		if (lock()) {
			rgb_color c = {255,0,0,255};
			v->SetFontAndColor(NULL, B_FONT_ALL, &c);
			unlock();
		}
		 		
		DBG( AWT, ("new KEdit %p (parent %p): %-30s\n", v, parent, txt)); 
	}
	
	BTextView * NATIVE() {
		return (BTextView *)getView();		
	}
	
	bool KEdit::OnChar( TCHAR c, int data) {
UNIMPLEMENTED((
		jstring aCmd;

		if (multi == false) {
			switch (c) {
			case VK_RETURN:
				GetWindowText( hwnd, X->buf, X->nBuf);
				aCmd = JniEnv->NewStringUTF( winString2UTF(X, X->buf, _tcslen(X->buf)) );
				X->jEvt = JniEnv->CallStaticObjectMethod( ActionEvent, getActionEvent,
													srcIdx, aCmd, 0 );
				return true;
			case VK_TAB:
				//non printable char bell
				return true;
			}
		}
		else {
			switch (c) {
			case VK_TAB:
				//do not process tabbing keys ( shift / ctrl )
				if ( getKeyMods() ){
					return true;
				}
				break;
			}
		}
));
		return false;
	}

	bool KEdit::OnKeyDown( int virtKey, int data) {
		KWnd::OnKeyDown( virtKey, data);
		//do not process tabbing keys ( shift / ctrl )
		return ( virtKey == VK_TAB);
	}

	void KEdit::finishResize() {
		DBG(AWT_GRA,("KEdit::finishResize\n"));
		BTextView * v = NATIVE();
		v->SetTextRect(v->Bounds());		
	}
	
	void KEdit::setEchoChar( jchar c) {
UNIMPLEMENTED((
		SendMessage( hwnd, EM_SETPASSWORDCHAR, (WPARAM) c, 0);
));
	}
	
	void KEdit::setCaretPos( int idx) {
		select( idx, idx);
	}
	
	void KEdit::select( int start, int end) {
UNIMPLEMENTED((
		SendMessage( hwnd, EM_SETSEL, (WPARAM)start, (LPARAM)end );
));
	}
	
	void KEdit::selectAll() {
		select( 0, -1);
	}
	
	void replaceSelection( TCHAR* txt) {
UNIMPLEMENTED((
		SendMessage( hwnd, EM_REPLACESEL, (WPARAM)TRUE, (LPARAM)txt);
));
	}

	void KEdit::setEditable( bool flag) {
		DBG(AWT, ("editSetEditable %p %d\n", this, flag));
		if (lock()) {
			NATIVE()->MakeEditable(flag);
			unlock();
		}
	}

	char * KEdit::getText() {
		char * txt = NULL;
		if (lock()) {
			txt = (char *) NATIVE()->Text();	/** @todo review cast */
			DBG(AWT, ("editGetText %p %10s\n", this, txt));
			unlock();
		}
		return txt;
	}
	
	void KEdit::setForeground(rgb_color c) {
		DBG(AWT, ("editSetForeground %p %x\n", this, c));
		NATIVE()->SetFontAndColor(NULL, B_FONT_ALL, &c);
	}
	
	void KEdit::setFont(BFont * f) {
		DBG(AWT, ("editSetFont %p %p\n", this, f));
		NATIVE()->SetFontAndColor(f, B_FONT_ALL, NULL);
	}
	
	void KEdit::setText( jstring jtext) {
		char * txt = java2CString( JniEnv, X, jtext);
		DBG(AWT, ("editSetText %p %10s\n", this, txt));
		if (lock()) {
			NATIVE()->SetText(txt);
			unlock();
		}
	}
};


/*************************
*	exported functions
**************************/
extern "C" {
	
	void* __cdecl
		Java_java_awt_Toolkit_editCreateField ( JNIEnv* env, jclass clazz, KWnd* parent, jstring txt )
	{
		return ( new KEdit( parent->getView(), java2CString( env, X, txt), 0 /* !!! ES_AUTOHSCROLL*/));
	}
	
	void* __cdecl
		Java_java_awt_Toolkit_editCreateArea ( JNIEnv* env, jclass clazz, KWnd* parent, jstring txt, jint style )
	{
		int ws = 0;
UNIMPLEMENTED((
		int ws = ES_MULTILINE | ES_AUTOVSCROLL;
		
		switch ( style ) {
		case ESB_VERT:
			ws |= WS_VSCROLL;
			break;
		case ESB_HORZ:
			ws |= WS_HSCROLL;
			break;
		case ESB_BOTH:
			ws |= WS_VSCROLL | WS_HSCROLL;
			break;
		}
));		
		return ( new KEdit( parent->getView(), java2CString( env, X, txt), ws) );
	}
	
	void __cdecl
		Java_java_awt_Toolkit_editSetEditable ( JNIEnv* env, jclass clazz, KEdit* wnd, jboolean edit )
	{
		wnd->setEditable(edit);
		// SendMessage( wnd->hwnd, EM_SETREADONLY, edit ? 0 : 1, 0);
	}
	
	jstring __cdecl
		Java_java_awt_Toolkit_editGetText ( JNIEnv* env, jclass clazz, KEdit* wnd)
	{
		/** @todo review UTF text conversion */
		char* str = wnd->getText();
		return (env->NewStringUTF( CString2UTF(X, str, strlen(str)) ));		
	}
	
	void __cdecl
		Java_java_awt_Toolkit_editSetText ( JNIEnv* env, jclass clazz, KEdit* wnd, jstring txt )
	{
		BRect r;			
		r = wnd->NATIVE()->TextRect();

		wnd->setText( txt);
	}
	
	jstring __cdecl
		Java_java_awt_Toolkit_editGetSelectedText ( JNIEnv* env, jclass clazz, KEdit* wnd)
	{
UNIMPLEMENTED((
#if !defined(UNDER_CE)
		jstring js;
		char* buf;
		int start, end;
		
		SendMessage( wnd->hwnd, EM_GETSEL, (WPARAM)&start, (LPARAM)&end);
		buf = ( char*)AWT_MALLOC( end-start);
		SendMessage( wnd->hwnd, EM_GETSELTEXT, 0, (LPARAM)buf);
		js = env->NewStringUTF( buf);
		AWT_FREE( buf);
		
		return js;
#else
		return (0);
#endif
));
	}
	
	jint __cdecl
		Java_java_awt_Toolkit_editGetSelectionStart ( JNIEnv* env, jclass clazz, KEdit* wnd)
	{
UNIMPLEMENTED((
		int idx = -1;
		SendMessage( wnd->hwnd, EM_GETSEL, (WPARAM)&idx, NULL);
		return idx;
));
	}
	
	jint __cdecl
		Java_java_awt_Toolkit_editGetSelectionEnd ( JNIEnv* env, jclass clazz, KEdit* wnd)
	{
UNIMPLEMENTED((
		int idx = -1;
		SendMessage( wnd->hwnd, EM_GETSEL, NULL, (LPARAM)&idx);
		return idx;
));
	}
	
	void __cdecl
		Java_java_awt_Toolkit_editSetSelectionStart ( JNIEnv* env, jclass clazz, KEdit* wnd, jint idx)
	{
UNIMPLEMENTED((
		int eIdx;
		SendMessage( wnd->hwnd, EM_GETSEL, NULL, (LPARAM)&eIdx);
		wnd->select( idx, eIdx);
));
	}
	
	void __cdecl
		Java_java_awt_Toolkit_editSetSelectionEnd ( JNIEnv* env, jclass clazz, KEdit* wnd, jint idx )
	{
UNIMPLEMENTED((
		int sIdx;
		SendMessage( wnd->hwnd, EM_GETSEL, (WPARAM)&sIdx, NULL);
		wnd->select( sIdx, idx);
));
	}
	
	void __cdecl
		Java_java_awt_Toolkit_editSelect ( JNIEnv* env, jclass clazz, KEdit* wnd, jstring txt, jint start, jint end )
	{
		wnd->select( start, end);
	}
	
	void __cdecl
		Java_java_awt_Toolkit_editSelectAll ( JNIEnv* env, jclass clazz, KEdit* wnd)
	{
		wnd->selectAll();
	}
	
	void __cdecl
		Java_java_awt_Toolkit_editSetCaretPosition ( JNIEnv* env, jclass clazz, KEdit* wnd,  jint idx)
	{
		wnd->setCaretPos( idx);
	}
	
	jint __cdecl
		Java_java_awt_Toolkit_editGetCaretPosition ( JNIEnv* env, jclass clazz, KEdit* wnd)
	{
		return -1;
	}
	
	void __cdecl
		Java_java_awt_Toolkit_editAppend ( JNIEnv* env, jclass clazz, KEdit* wnd, jstring txt )
	{
UNIMPLEMENTED((
		char* buf = java2CString( env, X, txt);
		wnd->setCaretPos( GetWindowTextLength( wnd->hwnd));
		wnd->replaceSelection( buf);
));
	}
	
	void __cdecl
		Java_java_awt_Toolkit_editInsert ( JNIEnv* env, jclass clazz, KEdit* wnd, jstring txt, jint idx )
	{
		char* buf = java2CString( env, X, txt);
		wnd->setCaretPos( idx);
		wnd->replaceSelection( buf);
	}
	
	void __cdecl
		Java_java_awt_Toolkit_editReplaceRange ( JNIEnv* env, jclass clazz, KEdit* wnd, jstring txt, jint start, jint end )
	{
		char* buf = java2CString( env, X, txt);
		wnd->select( start, end);
		wnd->replaceSelection( buf);
	}
	
	void __cdecl		
		Java_java_awt_Toolkit_editSetEchoChar ( JNIEnv* env, jclass clazz, KEdit* wnd, jchar c )		
	{
		wnd->setEchoChar( c);		
	}
	
}
