/**
 * @file gra.cpp
 * Graphics subsystem. 
 *
 * Copyright (c) 1999-2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * Portions Copyright (c) 1999
 *      Sim Ijskes
 *
 * @author	Tomaso Paoletti <tpaoletti@users.sourceforge.net>
 *          Sim IJskes <sim@nyx.xs4all.nl>
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "math.h"
#include "toolkit.hpp"

	/** Type of targets for the graInitGraphics function */
#define  WINDOW   0
#define  IMAGE    1
#define  GRAPHICS 2

	/** Flushes all cached graphical ops */
#define FLUSH(graphics)		(graphics->Flush())		

/** @todo review __cdecl */
#undef __cdecl
#define __cdecl

extern char * RootViewName;

/* !!!
extern "C" void getArcPoints( int x, int y, int width, int height,
						 int startAngle, int arcAngle, POINT& ps, POINT& pe) {
	
	int maxr, xc, yc;
	double rad, sa, ea;
	
	rad = -3.1415926535 / 180;
	maxr = ( max( width, height) ) /2;
	sa = rad * startAngle;
	ea = rad * (startAngle+arcAngle);
	xc = x + width/2 - 1;
	yc = y + height/2 - 1;
	
	ps.x = (int)( cos(sa) * maxr) + xc;
	ps.y = (int)( sin(sa) * maxr) + yc;
	pe.x = (int)( cos(ea) * maxr) + xc;
	pe.y = (int)( sin(ea) * maxr) + yc;
	
	//		printf( "x:%d y:%d w:%d h:%d x1:%d y1:%d x2:%d y2:%d\n",
	//			x, y, width, height, ps.x, ps.y, pe.x, pe.y);
}
*/

int __cdecl jarray2Points ( BPoint** pp, int x0, int y0,
						   jarray xPoints, jarray yPoints, int nPoints )
{
	register int i;
	int      n;
	jboolean isCopy;
	jint     *jx = JniEnv->GetIntArrayElements( xPoints, &isCopy);
	jint     *jy = JniEnv->GetIntArrayElements( yPoints, &isCopy);
	
	if ( (n = JniEnv->GetArrayLength( xPoints)) < nPoints ) nPoints = n;
	if ( (n = JniEnv->GetArrayLength( yPoints)) < nPoints ) nPoints = n;
	
	*pp = (BPoint*)getBuffer( X, (nPoints+1)*sizeof(BPoint));
	
	for ( i=0; i<nPoints; i++){
		(*pp)[i].x = jx[i] + x0;
		(*pp)[i].y = jy[i] + y0;
	}
	
	JniEnv->ReleaseIntArrayElements( xPoints, jx, JNI_ABORT);
	JniEnv->ReleaseIntArrayElements( yPoints, jy, JNI_ABORT);
	
	return nPoints;
}

//--------------------------------------------------------------------
/** The native part of a Java Graphics object 

    Unlike KWnd and friends, this class inherits directly from BView.
    Doing this we don't need an additional indirection on every call
    to the drawing functions. 

    See gra.cpp for the complete list of Java_* drawing functions.


	@todo add experimental buffering    
	@todo implement clipping (see function Graphics::setClip() )
*/
class Graphics : public BView {
public:	
	
	BRegion	oldreg;
	BRegion	newreg;

	KWnd	  *wnd;

	BPoint	cliOff;
#if CACHE_COLORS 
	rgb_color	fgClr;
	rgb_color	bgClr;
#endif

/* !!!
	HBITMAP	  hbm;
	HBRUSH	  brush;
	HPEN	  pen;
*/
	char      xor;
	rgb_color	  xclr;
	int		  x0;		/* !!! 32-bit integer */
	int		  y0;
	
		
	/** default constructor */
	Graphics() : BView(BRect(0,0,0,0), NULL, B_FOLLOW_ALL | B_FULL_UPDATE_ON_RESIZE, B_WILL_DRAW) {
	}
	

	/** 
		@todo ADD BUFFERING 
	
		EVERY GRAPHICAL OP LOCKS THE BView's LOOPER, CAUSING A MAIN 
   		BOTTLENECK THAT WE SHOULD ADDRESS... 	

   	Work to improve the graphics throughput:<BR> 
     	- a better locking scheme, with a 
       	  superLock/superUnlock pair wrapping complete paint operations 
     	- asynchronous calls to the Toolkit (protecting the shared Toolkit
data),         in most graphics ops we are thread-safe on our own,  
       so we don't need synchronized calls everywhere 
	*/ 
	bool		lock() {
		bool res;
		res = LockLooperWithTimeout(5e6) == B_OK;
		if (!res) printf("graphics: %p LOCK TIMEOUT (owner is %d)\n", 
			this, (int)Window()->LockingThread() );
		return res;
	}

	void		unlock() {
		UnlockLooper();
	}

	/** called by graInitGraphics, Graphics must already be locked */
	void init ( KWnd* pWnd, rgb_color cfg, rgb_color cbg, BFont* fnt) {

				
		setForeground(cfg);
		setBackground(cbg);	/* this sets only the B_SOLID_LOW color */
		SetViewColor(cbg);
		setFont(fnt);	
	
		xor = 0;
		wnd = pWnd;
		
/* !!!
		dc = hdc;
		hbm = hb;
		brush = CreateSolidBrush( fgClr);
		pen = CreatePen( PS_SOLID, 0, fgClr);

		DeleteObject( SelectObject( dc, brush));
		DeleteObject( SelectObject( dc, pen ));
		
		if ( hbm ) {
			SelectObject( dc, hbm );
		}
#if !defined(UNDER_CE)
		SetTextAlign( dc, TA_BASELINE | TA_LEFT);
#endif
		SetBkMode( dc, TRANSPARENT);
		SetTextColor( dc, fgClr);
*/		
		if ( wnd) {
			cliOff.x = wnd->insets.left;
			cliOff.y = wnd->insets.top;
		}
		else {
			cliOff.x = 0;
			cliOff.y = 0;
		}
	}
	
	void restoreBrush() {
/* !!!
		SelectObject( dc, brush);
*/	}
	
	void selectNullBrush() {
/* !!!
		SelectObject ( dc, HB_0);
*/	}
	
	void restorePen() {
/* !!!
		SelectObject( dc, pen);
*/	}
	
	void selectNullPen() {
/* !!!
		SelectObject ( dc, HP_0);
*/	}
	
	void setFont( BFont * fnt) {
		SetFont(fnt);
	}
	
/**	@todo setXORMode() is broken */
	void setXORMode( rgb_color xorClr) {
		xor = 1;
		xclr = xorClr;
		SetDrawingMode(B_OP_INVERT);
		SetLowColor( xclr );
#if CACHE_COLORS
		SetHighColor(fgClr);
#endif
	}
	
	void setPaintMode() {
		xor = 0;
/* 		SetLowColor(0xff, 0xff, 0xff);	 !!! why black color ? */
#if CACHE_COLORS
		SetHighColor(fgClr);
#endif
		SetDrawingMode(B_OP_COPY);	
	}
	
	bool setBackground ( rgb_color irgb) {
DBG(AWT_GRA, ("setBackground %x\n", irgb));
/* !!!
		if ( irgb < 0 )
			return false;
*/		
#if CACHE_COLORS
		bgClr = irgb;
		SetLowColor(bgClr);		// !!! for text drawing
#else
		SetViewColor(irgb);
		SetLowColor(irgb);		// !!! for text drawing		
#endif
		return true;
	}
	
	bool setForeground ( rgb_color irgb) {
/* !!!
		if ( irgb < 0 )
			return false;
*/		
#if CACHE_COLORS
		fgClr = irgb;
		SetHighColor(fgClr);
#else
		SetHighColor(irgb);
#endif				
		return true;
	}
	
	
	bool fillRect( int x, int y, int w, int h, rgb_color clr) {
		BRect r(x, y, x+w-1, y+h-1);	
		
/* !!!
		if ( clr < 0 )
			return false;
*/		

//		if ( *(int32 *)&clr != *(int32 *)&fgClr ) {
			rgb_color saveClr = HighColor();
			SetHighColor(clr);
			FillRect(r);
			SetHighColor(saveClr);
//		} else {
//			FillRect(r);
//         	}
		
		return true;
	}
	
	void drawRect( int x, int y, int w, int h ) {
		x+=x0;
		y+=x0;
		StrokeRect(BRect(x,y,x+w-1,y+h-1));
	}
	
	void drawLine ( int x1, int y1, int x2, int y2) {
		StrokeLine(BPoint(x1+x0,y1+y0),BPoint(x2+x0,y2+y0));
	}
	
	void drawText( int x, int y, TCHAR* txt, int len) {
		MovePenTo(x+x0, y+y0);
		DrawString(txt, len);
	}
	
	void translate( int x, int y) {
	/*
	* we use explicit translation, do not change
	* viewport here
		*/
		x0 = x - (int)cliOff.x;
		y0 = y - (int)cliOff.y;
	}
	
	void cleanUp() {
/* !!!		
		if ( brush ) { 
			SelectObject( dc, HB_0 );
			DeleteObject( brush);
			brush = 0;
		}
		if ( pen ) {
			SelectObject( dc, HP_0 );
			DeleteObject( pen);
			pen = 0;
		}
		if ( wnd ) {
			ReleaseDC( wnd->hwnd, dc);
			wnd = 0;
		}
		if ( hbm ) {
			hbm = 0;
		}
***/		
	}
	
	void setClip( int x, int y, int w, int h, bool set) {
		// TODO SGY see how BeOS manages references here
#if TODO

#ifdef DEBUG
		// for debugging purposes "scroll lock" disables clipping	
		if (modifiers() & B_SCROLL_LOCK) return;
#endif

		x += x0 ;
		y += y0 ;
    		newreg.Set(x,y,x+w,y+h);
	
		BRegion *region = new BRegion();
		
		/* !!! check this */
		if (set) {
			GetClippingRegion( region );
			Exclude( &oldreg );
			Include( &newreg );
		}
		ConstrainClippingRegion( region );
		oldreg = newreg;    
		delete region ;
#endif
	}
	
};


/*******************************************************************************
*	exported functions
*/


extern "C" {

		/** Initialize the native part of a Graphics object
		
		    Graphics objects are created from scratch or initialized from:
		    - a Window target
		    - an Image target
		    - another Graphics target
		    
		    THERE MAY STILL BE PROBLEMS WITH THE CASTS IN WHICH WE LOOSE
		    INFORMATION LIKE GRAPHICS OFFSETS AND CLIPPING COORDINATES
		    (WHY NOT EMBED ALL THESE INTO THE BVIEW ?)
		*/ 
	void* __cdecl
		Java_java_awt_Toolkit_graInitGraphics ( JNIEnv* env, jclass clazz,
		Graphics* gr, jobject tgt, jint tgtType,
		jint xOff, jint yOff,
		jint xClip, jint yClip, jint wClip, jint hClip,
		jobject fnt, jint fg, jint bg, jboolean blank )
	{
		KWnd *wnd = 0;
		Image *img = 0;
		Graphics *gr1;

		BBitmap * bmp = 0;

		DBG( AWT_GRA, ("initGraphics: %p (%p, %d) %d,%d "
				" %d,%d - %d,%d  %p %x %x\n",
				gr, tgt,tgtType, xOff,yOff, 
				xClip,yClip,wClip,hClip, fnt,fg,bg));

		if ( tgt ) {
			
			
			if ( gr ) {
				gr->cleanUp();
			}
			
			switch ( tgtType) {
			case WINDOW:	/* access KWnd's Graphics */
				wnd = (KWnd*)tgt;
				gr  = (Graphics *)wnd->getView();	/* !!! TYPE CAST IS BAD */
				break;
			case IMAGE:		/* access BBitmap's Graphics */
				img = (Image*)tgt;
				if (img) bmp = img->bitmap;
				if (bmp) gr  = (Graphics *)bmp->FindView(RootViewName);	/* !!! TYPE CAST IS BAD */
				break;
			case GRAPHICS:	/* clone the target Graphics */
				gr1 = (Graphics*)tgt;
				wnd = gr1->wnd;
				gr = (Graphics *)wnd->getView();	/* !!! TYPE CAST IS BAD */
/* !!!
				if ( wnd ) {
					hdc = GetDC( wnd->hwnd);
				}
				else {
					hdc = gr1->dc;
				}
				hbm = gr1->hbm;
*/
				break;
			}
		}
		else if ( gr ) {
			wnd = gr->wnd;	/* use the same Graphics */
/* !!!
			hdc = gr->dc;
			hbm = gr->hbm;
*/
		}
		
		if ( ! gr) {
			gr = new Graphics();
		}
		
		if (gr->lock()) {
			
			gr->init( wnd, RGBCAST(fg), RGBCAST(bg), (BFont*)fnt);
			gr->translate( xOff, yOff);
			gr->setClip( xClip, yClip, wClip, hClip, TRUE);

			if ( blank) {
				gr->fillRect( xClip, yClip, wClip, hClip, RGBCAST(bg));
			}
			gr->unlock();
			FLUSH(gr);
			return gr;
		} else 
			return NULL;		// !!! we leak memory
	}
	
	
	void __cdecl
		Java_java_awt_Toolkit_graFreeGraphics ( JNIEnv* env, jclass clazz, Graphics* gr )
	{
		DBG( AWT_GRA, ("freeGraphics: %p\n", gr));
		
		gr->cleanUp();
		/* !!! we don't own it, the owner of the BView will dispose for us
		delete gr; 
		*/
	}
	
	
	void __cdecl
		Java_java_awt_Toolkit_graCopyArea ( JNIEnv* env, jclass clazz, Graphics* gr, 
		jint x, jint y, jint width, jint height,
		jint xDelta, jint yDelta )
	{
		DBG( AWT_GRA, ("copyArea: %p, %d,%d-%d,%d  %d,%d\n", gr, x,y,width,height, xDelta,yDelta));
		if (gr->lock()) {
			x += gr->x0; y += gr->y0;
			BRect src(x,y,x+width,y+height);  
    		gr->CopyBits( src, src.OffsetBySelf(BPoint(xDelta,yDelta)) );
			gr->unlock();
		}
		FLUSH(gr);
	}
	
	void __cdecl
		Java_java_awt_Toolkit_graClearRect ( JNIEnv* env, jclass clazz, Graphics* gr,
		jint x, jint y, jint width, jint height )
	{
		DBG( AWT_GRA, ("clearRect: %p, %d,%d-%d,%d %x\n", gr, x,y,width,height, 0 /*gr->bgClr*/));
		if (gr->lock()) {
#if CACHE_COLORS
			gr->fillRect( x, y, width, height, gr->bgClr);	
#else
			gr->fillRect( x, y, width, height, gr->ViewColor());	
#endif
			gr->unlock();
		}
		FLUSH(gr);
	}
	
	
	void __cdecl
		Java_java_awt_Toolkit_graDrawBytes ( JNIEnv* env, jclass clazz, Graphics* gr,
		jarray jBytes, jint offset, jint len, jint x, jint y )
	{
		jboolean isCopy;
		int      n;
		jbyte    *jb, *jbo;
		char	   *cs;
		
		if ( !jBytes ) return;
		
		n  = env->GetArrayLength( jBytes);
		jb = env->GetByteArrayElements( jBytes, &isCopy);
		jbo = jb + offset;
		
		if ( offset+len > n )
			len = n - offset;
		
		cs = jbyte2CString( X, jbo, len);
		
		DBG( AWT_GRA, ("drawBytes: %p, %p,%d,%d  \"%s\"  %d,%d\n", gr, jBytes,offset,len, cs, x,y));
		gr->drawText( x, y, cs, len);
		
		env->ReleaseByteArrayElements( jBytes, jb, JNI_ABORT);
		FLUSH(gr);
	}
	
	
	void __cdecl
		Java_java_awt_Toolkit_graDrawChars ( JNIEnv* env, jclass clazz, Graphics* gr,
		jarray jChars, jint offset, jint len, jint x, jint y )
	{
		jboolean isCopy;
		int      n;
		jchar    *jc, *jco;
		TCHAR	 *cs;
		
		if ( !jChars ) return;
		
		n   = env->GetArrayLength( jChars);
		jc = env->GetCharArrayElements( jChars, &isCopy);
		jco = jc + offset;
		
		if ( offset+len > n )
			len = n - offset;
		
		cs = jchar2CString( X, jco, len);
		
		DBG( AWT_GRA, ("drawChars: %p, %p,%d,%d  \"%s\"  %d,%d\n", gr, jChars,offset,len, cs, x,y));
		if (gr->lock()) {
			gr->drawText( x, y, cs, len);
			gr->unlock();
		}
		env->ReleaseCharArrayElements( jChars, jc, JNI_ABORT);
		FLUSH(gr);
	}
	
	
	void __cdecl
		Java_java_awt_Toolkit_graDrawString ( JNIEnv* env, jclass clazz,
		Graphics* gr, jstring str, jint x, jint y )
	{
		char  *jc = java2CString( env, X, str);
		
		DBG( AWT_GRA, ("drawString: %p  \"%s\"  %d,%d\n", gr, jc, x,y));
		if (gr->lock()) {
			gr->drawText( x, y, jc, strlen( jc));
			gr->unlock();
		}
		FLUSH(gr);
	}
	
	
	void __cdecl
		Java_java_awt_Toolkit_graDrawLine ( JNIEnv* env, jclass clazz, Graphics* gr,
		jint x1, jint y1, jint x2, jint y2 )
	{
		DBG( AWT_GRA, ("drawLine: %p, %d,%d - %d,%d\n", gr, x1,y1, x2,y2));
		if (gr->lock()) {
			gr->drawLine( x1, y1, x2, y2);
			gr->unlock();
		}
		FLUSH(gr);
	}
	
	void __cdecl
		Java_java_awt_Toolkit_graDrawArc ( JNIEnv* env, jclass clazz, Graphics* gr,
		jint x, jint y, jint width, jint height,
		jint startAngle, jint arcAngle )
	{
		DBG( AWT_GRA, ("drawArc: %p, %d,%d-%d,%d  %d,%d\n", gr,
				x,y,width,height, startAngle,arcAngle));
		if (gr->lock()) {
			x += gr->x0;
			y += gr->y0;
			gr->StrokeArc(BRect(x, y, x+width, y+height), startAngle, arcAngle);
			gr->unlock();
		}
		FLUSH(gr);
	}	
	
	void __cdecl
		Java_java_awt_Toolkit_graFillArc ( JNIEnv* env, jclass clazz, Graphics* gr,
		jint x, jint y, jint width, jint height,
		jint startAngle, jint arcAngle )
	{
  		DBG( AWT_GRA, ("fillArc: %p, %d,%d-%d,%d  %d,%d\n", gr,
				x,y,width,height, startAngle, arcAngle));
		if (gr->lock()) {
			x += gr->x0;
			y += gr->y0;
			gr->FillArc(BRect(x, y, x+width, y+height), startAngle, arcAngle);
      		gr->unlock();
		}
		FLUSH(gr);
	}
	
	void __cdecl
		Java_java_awt_Toolkit_graDrawOval ( JNIEnv* env, jclass clazz, Graphics* gr,
		jint x, jint y, jint width, jint height )
	{
		DBG( AWT_GRA, ("drawOval: %p, %d,%d - %d,%d\n", gr, x,y,width,height));
		if (gr->lock()) {
			x += gr->x0;
			y += gr->y0;
			gr->StrokeArc(BRect(x, y, x+width, y+height), 0, 360);
			gr->unlock();
		}
		FLUSH(gr);
	}
	
	
	void __cdecl
		Java_java_awt_Toolkit_graFillOval ( JNIEnv* env, jclass clazz, Graphics* gr,
		jint x, jint y, jint width, jint height )
	{
		DBG( AWT_GRA, ("fillOval: %p, %d,%d - %d,%d\n", gr, x,y,width,height));
		if (gr->lock()) {
			x += gr->x0;
			y += gr->y0;
			gr->FillArc(BRect(x, y, x+width, y+height), 0, 360);
			gr->unlock();
		}
		FLUSH(gr);
	}
	
	
	void __cdecl
		Java_java_awt_Toolkit_graDrawPolygon ( JNIEnv* env, jclass clazz, Graphics* gr,
		jarray xPoints, jarray yPoints, jint nPoints )
	{
		BPoint *p;
		
		DBG( AWT_GRA, ("drawPolygon: %p, %p,%p  %d\n", gr, xPoints, yPoints, nPoints));
		
		if ( !xPoints || !yPoints ) return;
		
		nPoints = jarray2Points( &p, gr->x0, gr->y0, xPoints, yPoints, nPoints);
		
		if (gr->lock()) {   
			gr->StrokePolygon(p, nPoints, true);
			gr->unlock();
		}		
		FLUSH(gr);  
	}
	
	
	void __cdecl
		Java_java_awt_Toolkit_graDrawPolyline ( JNIEnv* env, jclass clazz, Graphics* gr,
		jarray xPoints, jarray yPoints, jint nPoints )
	{
		BPoint *p;
		DBG( AWT_GRA, ("drawPolyline: %p, %p,%p  %d\n", gr, xPoints, yPoints, nPoints));

		nPoints = jarray2Points( &p, gr->x0, gr->y0, xPoints, yPoints, nPoints);

		if (gr->lock()) {
			gr->StrokePolygon(p, nPoints, false);
			gr->unlock();	
		}
		FLUSH(gr);
	}
	
	
	void __cdecl
		Java_java_awt_Toolkit_graFillPolygon ( JNIEnv* env, jclass clazz, Graphics* gr,
		jarray xPoints, jarray yPoints, jint nPoints )
	{
		BPoint *p;		
		DBG( AWT_GRA, ("fillPolygon: %p, %p,%p  %d\n", gr, xPoints, yPoints, nPoints));
		
		if ( !xPoints || !yPoints ) return;

		nPoints = jarray2Points( &p, gr->x0, gr->y0, xPoints, yPoints, nPoints);

		if (gr->lock())  {
			gr->FillPolygon(p, nPoints);
			gr->unlock();
		}	
		FLUSH(gr);
	}
	
	void __cdecl
		Java_java_awt_Toolkit_graDrawRect ( JNIEnv* env, jclass clazz, Graphics* gr,
		jint x, jint y, jint width, jint height )
	{
		DBG( AWT_GRA, ("drawRect: %p, %d,%d - %d,%d\n", gr, x,y,width,height));
		
		if ( (width >= 0) && (height >= 0) && gr->lock()){
			gr->drawRect( x, y, width, height);
			gr->unlock();
		}
		FLUSH(gr);
	}
	
	
	void __cdecl
		Java_java_awt_Toolkit_graFillRect ( JNIEnv* env, jclass clazz, Graphics* gr,
		jint x, jint y, jint width, jint height )
	{
		DBG( AWT_GRA, ("fillRect: %p, %d,%d - %d,%d %x\n", gr, x,y,width,height, 0 /*gr->fgClr*/));
		if (gr->lock()) {
#if CACHE_COLORS
			gr->fillRect( x, y, width, height, gr->fgClr);
#else
			gr->fillRect( x, y, width, height, gr->HighColor());
#endif
			gr->unlock();
		}
		FLUSH(gr);
	}
	
	
	void __cdecl
		Java_java_awt_Toolkit_graDrawRoundRect ( JNIEnv* env, jclass clazz, Graphics* gr,
		jint x, jint y, jint width, jint height, 
		jint wArc, jint hArc )
	{
		
		DBG( AWT_GRA, ("drawRoundRect: %p, %d,%d - %d,%d  %d,%d\n", gr,
			x,y,width,height, wArc, hArc));
		
		if (gr->lock()) {
			x+=gr->x0;
			y+=gr->y0;
			gr->StrokeRoundRect(BRect(x, y, x+width-1, y+height-1), wArc, hArc);
			gr->unlock();
		}
		FLUSH(gr);
	}
	
	void __cdecl
		Java_java_awt_Toolkit_graFillRoundRect ( JNIEnv* env, jclass clazz, Graphics* gr,
		jint x, jint y, jint width, jint height, 
		jint wArc, jint hArc )
	{
		
		DBG( AWT_GRA, ("fillRoundRect: %p, %d,%d - %d,%d  %d,%d\n", gr,
			x,y,width,height, wArc, hArc));
		
		if (gr->lock()) {
			x+=gr->x0;
			y+=gr->y0;
			gr->FillRoundRect(BRect(x, y, x+width-1, y+height-1), wArc, hArc);
			gr->unlock();
		}
		FLUSH(gr);
	}
	
	
	void __cdecl
		Java_java_awt_Toolkit_graDraw3DRect ( JNIEnv* env, jclass clazz, Graphics* gr,
		jint x, jint y, jint width, jint height,
		jboolean raised, jint javargb )
	{
		rgb_color   bright, dark;
		float		xw, yh;	
		
		DBG( AWT_GRA, ("draw3DRect: %p, %d,%d - %d,%d  %d %x\n", gr,
			x,y,width,height, raised, javargb));
		
		// !!! CHECK TWICE, ColorJ2N SHOULDN'T BE NECESSARY
		/* we should use the same mechanism like ordinary Graphics brighter()/darker() here */
		dark   = ColorJ2N((int)Java_java_awt_Toolkit_clrDark( env, clazz, javargb));
		bright = ColorJ2N((int)Java_java_awt_Toolkit_clrBright( env, clazz, javargb));
		/* Set
		pd = CreatePen( PS_SOLID, 0, dark);
		pb = CreatePen( PS_SOLID, 0, bright);
		*/

		if (gr->lock()) {
			gr->PushState();

			xw = x + width;
			yh = y + height;

			gr->SetHighColor(raised ? bright : dark);
			gr->drawLine( x, y, xw-1, y);
			gr->drawLine( x, y, x, yh);
		
			gr->SetHighColor(raised ? dark : bright);
			gr->drawLine( x+1, yh, xw, yh);
			gr->drawLine( xw, y, xw, yh);
		
			gr->PopState();		
			gr->unlock();
		}
		FLUSH(gr);
	}
	
	void __cdecl
		Java_java_awt_Toolkit_graFill3DRect ( JNIEnv* env, jclass clazz, Graphics* gr,
		jint x, jint y, jint width, jint height,
		jboolean raised, jint javargb )
	{
		
		DBG( AWT_GRA, ("fill3DRect: %p, %d,%d - %d,%d  %d %x\n", gr,
			x,y,width,height, raised, javargb));	
		if (gr->lock()) {
			gr->fillRect( x+1, y+1, width-2, height-2, ColorJ2N(javargb));
			gr->unlock();
		}
		Java_java_awt_Toolkit_graDraw3DRect( env, clazz, gr, x, y, width-1, height-1, raised, javargb);
	}
	
	
	void __cdecl
		Java_java_awt_Toolkit_graAddClip ( JNIEnv* env, jclass clazz, Graphics* gr,
		jint xClip, jint yClip, jint wClip, jint hClip )
	{
		DBG( AWT_GRA, ("addClip: %p, %d,%d - %d,%d\n", gr, xClip, yClip, wClip, hClip));
		gr->setClip( xClip, yClip, wClip, hClip, FALSE);	  
	}
	
	
	void __cdecl
		Java_java_awt_Toolkit_graSetClip ( JNIEnv* env, jclass clazz, Graphics* gr,
		jint xClip, jint yClip, jint wClip, jint hClip )
	{
		DBG( AWT_GRA, ("setClip: %p, %d,%d - %d,%d\n", gr, xClip, yClip, wClip, hClip));
		gr->setClip( xClip, yClip, wClip, hClip, TRUE);	  
	}
	
	void __cdecl
		Java_java_awt_Toolkit_graSetColor ( JNIEnv* env, jclass clazz, Graphics* gr, jint clr )
	{
		DBG( AWT_GRA, ("setColor: %p, %x\n", gr, clr));
		
		if (gr->lock()) {
/* !!!
		if ( gr->xor )
			clr ^= gr->xclr;
*/	
			gr->setForeground( RGBCAST(clr));
			gr->unlock();
		}
		FLUSH(gr);
	}
	
	void __cdecl
		Java_java_awt_Toolkit_graSetBackColor ( JNIEnv* env, jclass clazz, Graphics* gr, jint clr )
	{
		DBG( AWT_GRA, ("setBackColor: %p, %x\n", gr, clr));
		if (gr->lock()) {
			gr->setBackground( RGBCAST(clr));
			gr->unlock();
		}
		FLUSH(gr);
	}
	
	void __cdecl
		Java_java_awt_Toolkit_graSetFont ( JNIEnv* env, jclass clazz, Graphics* gr, jobject fnt )
	{
		DBG( AWT_GRA, ("setFont: %p, %p\n", gr, fnt));
		if (gr->lock()) {
			gr->setFont( (BFont *)fnt);
			gr->unlock();
		}
	}
	
	void __cdecl
		Java_java_awt_Toolkit_graSetOffset ( JNIEnv* env, jclass clazz, Graphics* gr, jint xOff, jint yOff )
	{
		DBG( AWT_GRA, ("setOffset: %p, %d,%d\n", gr, xOff, yOff));
		if (gr->lock()) {
			gr->translate( xOff, yOff);
			gr->unlock();
		}
	}
	
	void __cdecl
		Java_java_awt_Toolkit_graSetPaintMode ( JNIEnv* env, jclass clazz, Graphics* gr )
	{
		DBG( AWT_GRA, ("setPaintMode: %p\n", gr));
		if( gr->lock() ) {
			gr->setPaintMode();
			gr->unlock();
		}
	}
	
	void __cdecl
		Java_java_awt_Toolkit_graSetXORMode ( JNIEnv* env, jclass clazz, Graphics* gr, jint xorClr )
	{
		DBG( AWT_GRA, ("setXORMode: %p, %x\n", gr, xorClr));
TOFIX();
		if( gr->lock() ) {
			gr->setXORMode(RGBCAST(xorClr));
			gr->unlock();
		}
	}
	
	void __cdecl
		Java_java_awt_Toolkit_graSetVisible ( JNIEnv* env, jclass clazz, Graphics* gr, jint isVisible )
	{
		DBG( AWT_GRA, ("setVisble: %p  %d\n", gr, isVisible));
		
		/*
		* This is rather a hack to "defuse" a Graphics object, but we don't want to
		* add checks on every draw op to test if the target is invisible
		*/
		
		if ( !isVisible ) {
			if ( gr->y0 >= -10000 ) 
				gr->y0 -= 100000;
		}
		else {
			if ( gr->y0 <= -10000 )
				gr->y0 += 100000;
		}
	}
	
	
	/************************************************************************************
	* image rendering
	*/
	
	void __cdecl
		drawAlphaImage ( Graphics* gr, Image* img,
		jint srcX, jint srcY, jint dstX, jint dstY,
		jint width, jint height, jint bgval )
	{
		DBG(AWT_IMG, ("drawAlphaImage: %p %p (%p,%x,%x %d,%d) %d,%d, %d,%d, %d,%d, %x\n",
				 gr, img, img->bitmap, 0 /*img->xMask*/,0 /*img->alpha*/,  img->width, img->height,
				 srcX,srcY,  dstX,dstY,	 width,height, bgval));
UNIMPLEMENTED((
		// !!! SOURCE srcX,srcY IS NOT USED - FIX
));		
	}
	
	
	void __cdecl
		Java_java_awt_Toolkit_graDrawImage ( JNIEnv* env, jclass clazz, Graphics* gr, Image* img,
		jint srcX, jint srcY,
		jint dstX, jint dstY,
		jint width, jint height, jint bgval )
	{
		DBG(AWT_IMG, ("drawImage: %p %p (%p,%x,%x %d,%d) %d,%d, %d,%d, %d,%d, %x\n",
				 gr, img, img->bitmap, 0 /*img->xMask*/,0 /*img->alpha*/,  img->width, img->height,
				 srcX,srcY,  dstX,dstY,	 width,height, bgval));
		
UNIMPLEMENTED((
		// !!! SOURCE srcX,srcY IS NOT USED - FIX
));		
		dstX += gr->x0;
		dstY += gr->y0;

		if (img && gr->lock())
		{   
			//drawing_mode saveMode = gr->DrawingMode();
			gr->PushState();
			
/* !!!
			if ( bgval != -1 ) {
				gr->SetHighColor(RGBCAST(bgval));
				gr->FillRect( BRect(dstX, dstY, dstX+width, dstY+height));
			}
*/

			if (img->trans<0) {
      			gr->SetDrawingMode(B_OP_COPY);
			} else {
				gr->SetDrawingMode(B_OP_ALPHA);
				gr->SetBlendingMode(B_PIXEL_ALPHA, B_ALPHA_OVERLAY);
			}	
			gr->DrawBitmap(img->bitmap, BPoint(dstX, dstY));

			//gr->SetDrawingMode(saveMode);
			gr->PopState();
			
			gr->unlock();
		}
		FLUSH(gr); 
	}
	
	
	void __cdecl
		Java_java_awt_Toolkit_graDrawImageScaled ( JNIEnv* env, jclass clazz, Graphics* gr, Image* img,
		jint dx0, jint dy0, jint dx1, jint dy1,
		jint sx0, jint sy0, jint sx1, jint sy1, jint bgval )
	{
		DBG(AWT_IMG, ("drawImageScaled: %p %p (%x), %d,%d,%d,%d, %d,%d,%d,%d, %x\n",
				 gr, img, 0/*img->xImg*/, dx0, dy0, dx1, dy1, sx0, sy0, sx1, sy1, bgval));
		dx0 += gr->x0;
		dx1 += gr->x0;
		dy0 += gr->y0;
		dy1 += gr->y0;

		if (img && gr->lock()) {
			BRect dest(dx0,dy0,dx1,dy1);
			//drawing_mode saveMode = gr->DrawingMode();
			gr->PushState();
			
/* !!!
			if ( bgval != -1 ) {
				gr->SetHighColor(RGBCAST(bgval));
				gr->FillRect(dest);
			}
*/
			if (img->trans<0) {
				gr->SetDrawingMode(B_OP_COPY);
			} else {
				gr->SetDrawingMode(B_OP_ALPHA);
				gr->SetBlendingMode(B_PIXEL_ALPHA, B_ALPHA_OVERLAY);
      		}	
      		gr->DrawBitmap(img->bitmap, BRect(sx0,sy0,sx1,sy1), dest);
      		//gr->SetDrawingMode(saveMode);
			gr->PopState();
			gr->unlock();
  		}
  		FLUSH(gr);   
	}
}
