/**
 * @file lst.cpp 
 * List classes
 *
 * Copyright (c) 1998-2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * @author T. Paoletti <tpaoletti@users.sourceforge.net>,
 *         <your name here>
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */
 
#include "toolkit.hpp"
#include "cmnwnd.hpp"

/** 
	Native list widget.
*/
class KList:public KWnd {
public:		
	jboolean freezed;
	jboolean multiSel;
	BListView *fList;
	
	KList( BView * parent, jboolean multi) {
/*
		int style = multi ? LBS_MULTIPLESEL : 0;
		style |= DEF_WND_STYLE | WS_VSCROLL | WS_HSCROLL | LBS_NOTIFY | LBS_NOINTEGRALHEIGHT;
		hwnd = CreateWindowEx( WS_EX_CLIENTEDGE, TEXT("LISTBOX"), NULL, style, 0, 0, 0, 0,
			parent, (HMENU)0, NULL, NULL);
		defWndProc = (WNDPROC)SetWindowLong( hwnd, GWL_WNDPROC, (LONG)WndProc);
		freezed = false;
		multiSel = multi;
*/
		BRect r(0,0,0,0);
		
		fList = new BListView(r, "KList", 
			(multi ? B_MULTIPLE_SELECTION_LIST : B_SINGLE_SELECTION_LIST),
			B_FOLLOW_ALL_SIDES);
		freezed = false;
		multiSel = multi;

#if SCROLL
		BScrollView * scroll = new BScrollView("KList,scroll", fList, 
									B_FOLLOW_ALL_SIDES, 0, true, true );
		
		parent->AddChild(scroll);
		setView(scroll);				// !!! scroll or fList, which one ?
#else
		parent->AddChild(fList);
		setView(fList);
#endif
	}
	
	/** Access to widget's view.
	
		A note: we have two BViews into one: a BListView and a
		container BScrollView from outside (the KWnd hierarchy).  
		We want to handle the scroll view properties but from
		inside 	(this module) we want to work on the list (fList).
		
		See btn.cpp for the different NATIVE() access method */   
	BListView * KList::NATIVE() {
	  	
		return (BListView *)getView();
		// return fList;
	}
 	
 	

	void KList::setMode( bool multi ) {
  	
		list_view_type ltype = multi ? B_MULTIPLE_SELECTION_LIST :
B_SINGLE_SELECTION_LIST;
		NATIVE()->SetListType( ltype);
	}
		
 
	bool KList::OnCommand( int cmd) {
 
	UNIMPLEMENTED((
		jstring aCmd;
		int idx, cmdId;
		TCHAR* str;
		
		switch ( cmd) {
		case LBN_SELCHANGE:
			//multiple mode gets focused item
			idx = getSelectionIdx();
			if (  multiSel && ! isSelected( idx)) {
				cmdId = DESELECTED;
			}
			else {
				cmdId = SELECTED;
			}
			JniEnv->CallStaticObjectMethod( ItemEvent, getItemEvent,
				srcIdx, cmdId, idx );
			break;
		case LBN_DBLCLK:
			str = getSelectionString();
			aCmd = JniEnv->NewStringUTF( winString2UTF(X, str, _tcslen(str)) );
			JniEnv->CallStaticObjectMethod( ActionEvent, getActionEvent,
				srcIdx, aCmd, 0 );
			break;
		case LBN_SELCANCEL:
			break;
		case LBN_SETFOCUS:
			break;
		case LBN_KILLFOCUS:
			break;
		case LBN_ERRSPACE:
			break;
		}
));		
		return TRUE;
	}
	
	char* KList::getSelectionString() {
UNIMPLEMENTED((
		int idx = SendMessage( hwnd, LB_GETCURSEL, 0, 0);
		if ( idx != LB_ERR ) {
			SendMessage( hwnd, LB_GETTEXT, (WPARAM)idx, (LPARAM)X->buf);
			return X->buf;
		}
		else {
			return (TCHAR*)0;
		}
));
		return (char*)0;
	}
	
	bool KList::isSelected( int idx) {
/*
		return (SendMessage( hwnd, LB_GETSEL, (WPARAM)idx, 0) > 0);
*/
		bool sel = false;
		if (lock()) {
			sel = NATIVE()->IsItemSelected(idx);
			unlock();
		}
		return sel;
	}
	
	int KList::getSelectionIdx() {
UNIMPLEMENTED((
		//for single selection lists
		return (int)SendMessage( hwnd, LB_GETCURSEL, 0, 0);
));
		int sel = 0;
		if (lock()) {
			sel = NATIVE()->CurrentSelection(0);
			unlock();
		}
		return sel;
	}
	
	jobject KList::getSelectionIdxs() {
/*		//for multiple-selection lists
		jboolean	isCopy;
		jarray		selArray;
		jint*		sels;
		int			numSel = (int)SendMessage( hwnd, LB_GETSELCOUNT, 0, 0);
		
		if ( numSel <= 0 )
			return NULL;
		
		selArray  = JniEnv->NewIntArray( numSel);
		sels = JniEnv->GetIntArrayElements( selArray, &isCopy);
		
		SendMessage( hwnd, LB_GETSELITEMS, (WPARAM)numSel, (LPARAM)sels);
		
		if ( isCopy ) {
			JniEnv->ReleaseIntArrayElements( selArray, sels, JNI_COMMIT);
		}
		
		return selArray;
*/
		//for multiple-selection lists
		jboolean	isCopy;
		jarray		selArray;
		jint*		sels;
		int32		numSel;
		
		numSel = 0;
UNIMPLEMENTED((
		BListView * list = NATIVE();
		while (list->CurrentSelection(numSel) >=0) {
			numSel++;
		}
));
		selArray  = JniEnv->NewIntArray( numSel);
		sels = JniEnv->GetIntArrayElements( selArray, &isCopy);
		
UNIMPLEMENTED((
		jint		csel;
		int32		n;
		n = 0;
		while ((csel = list->CurrentSelection(n)) >= 0) {
			sels[n++] = csel;
		}
));
		
		if ( isCopy ) {
			JniEnv->ReleaseIntArrayElements( selArray, sels, JNI_COMMIT);
		}

		return selArray;
	}
	
	bool KList::updateHScroll() {
UNIMPLEMENTED((
		
		HDC hdc;
		SIZE sz;
		int idx, len, maxLen;
		
		if ( freezed ) {
			return FALSE;
		}
		
		maxLen = 0;
		hdc = GetDC( hwnd);
		
		for ( idx=0;;idx++) {
			len = (int)SendMessage( hwnd, LB_GETTEXT, (WPARAM)idx, (LPARAM)X->buf);
			if ( len == LB_ERR )
				break;
			GetTextExtentPoint32( hdc, X->buf, len, &sz);
			if ( sz.cx > maxLen ) {
				maxLen = sz.cx;
			}
		}
		
		ReleaseDC( hwnd, hdc);
		
		maxLen -= GetSystemMetrics( SM_CXVSCROLL);
		
		SendMessage( hwnd, LB_SETHORIZONTALEXTENT, (WPARAM)maxLen, 0);
));

		// !!! IMPLEMENT		
		return TRUE;
	}
	};
	
	
	
	/*************************
	*	exported functions
	**************************/
	extern "C" {
		
		void* __cdecl
			Java_java_awt_Toolkit_lstCreateList ( JNIEnv* env, jclass clazz, KWnd* parent, jboolean multi)
		{
			return ( new KList( parent->getView(), multi) );
		}
		
		void __cdecl
			Java_java_awt_Toolkit_lstAppendItem ( JNIEnv* env, jclass clazz, KList* wnd, jstring item)
		{
/*
			TCHAR* buf = java2WinString( env, X, item);
			SendMessage( wnd->hwnd, LB_ADDSTRING, 0, (LPARAM)buf);
			wnd->updateHScroll();
*/
			char* buf = java2CString( env, X, item);
			DBG(AWT, ("lstAppendItem %p : %s\n", wnd, buf)); 
			if (wnd->lock()) {
				wnd->NATIVE()->AddItem(new BStringItem(buf));					
				wnd->unlock();
			}
			wnd->updateHScroll();	// !!! do we need this ?
		}
		
		jint __cdecl
			Java_java_awt_Toolkit_lstGetVisibleIndex ( JNIEnv* env, jclass clazz, KList* wnd)
		{
UNIMPLEMENTED((
			return (jint)SendMessage( wnd->hwnd, LB_GETTOPINDEX, 0, 0);
));
			return 0;
		}
		
		jobject __cdecl
			Java_java_awt_Toolkit_lstGetSelectionIdxs ( JNIEnv* env, jclass clazz, KList* wnd)
		{
			jobject sels = NULL;
			if (wnd->lock()) {
				sels = wnd->getSelectionIdxs();
				wnd->unlock();
			}
			return sels;
		}
		
		void __cdecl
			Java_java_awt_Toolkit_lstInsertItem ( JNIEnv* env, jclass clazz, KList* wnd, jstring item, jint idx)
		{
/*
			char* buf = java2CString( env, X, item);
			SendMessage( wnd->hwnd, LB_INSERTSTRING, idx, (LPARAM)buf);
*/
			char* buf = java2CString( env, X, item);
			DBG(AWT, ("lstInsertItem %p %d : %s\n", wnd, idx, buf)); 
			if (wnd->lock()) {
				wnd->NATIVE()->AddItem(new BStringItem(buf), idx);					
				wnd->unlock();
			}
			wnd->updateHScroll();	// !!! do we need this ?
		}
		
		void __cdecl
			Java_java_awt_Toolkit_lstFreeze ( JNIEnv* env, jclass clazz, KList* wnd)
		{
UNIMPLEMENTED((
			SendMessage( wnd->hwnd, WM_SETREDRAW, (WPARAM) FALSE, 0);
			wnd->freezed = TRUE;
));
		}
		
		void __cdecl
			Java_java_awt_Toolkit_lstThaw ( JNIEnv* env, jclass clazz, KList* wnd)
		{
UNIMPLEMENTED((
			SendMessage( wnd->hwnd, WM_SETREDRAW, (WPARAM) TRUE, 0);
			wnd->freezed = FALSE;
			wnd->updateHScroll();
));		}
		
		void __cdecl
			Java_java_awt_Toolkit_lstUnselectItem ( JNIEnv* env, jclass clazz, KList* wnd, jint idx)
		{
UNIMPLEMENTED((
			if ( wnd->multiSel ) {
				SendMessage( wnd->hwnd, LB_SETSEL, (WPARAM)false, (LPARAM)idx);
			}
			else if ( wnd->getSelectionIdx() == idx ) {
				//remove current selection
				SendMessage( wnd->hwnd, LB_SETCURSEL, (WPARAM)-1, 0);
			}
));
			if (wnd->lock()) {
				if ( wnd->multiSel) {
					wnd->NATIVE()->DeselectAll();
				} else {
					wnd->NATIVE()->Deselect(idx);
				}
				wnd->unlock();
			}
			
		}
		
		void __cdecl
			Java_java_awt_Toolkit_lstSelectItem ( JNIEnv* env, jclass clazz, KList* wnd, jint idx)
		{
/*
			if ( wnd->multiSel ) {
				SendMessage( wnd->hwnd, LB_SETSEL, (WPARAM)true, (LPARAM)idx);
			}
			else {
				SendMessage( wnd->hwnd, LB_SETCURSEL, (WPARAM)idx, 0);
			}
*/
			if (wnd->lock()) {
				if ( wnd->multiSel) {
					wnd->NATIVE()->Select(idx, true);	// extend selection
				} else {
					wnd->NATIVE()->Select(idx);
				}
				wnd->unlock();
			}
		}
		
		void __cdecl
			Java_java_awt_Toolkit_lstRemoveAll ( JNIEnv* env, jclass clazz, KList* wnd)
		{
UNIMPLEMENTED((
			SendMessage( wnd->hwnd, LB_RESETCONTENT, 0, 0);
			wnd->updateHScroll();
));
/* 
			BListView   *list = NATIVE();
			
			int32 		i;
			int32 		nitems = list->CountItems();
			BListItem   *items = list->Items();
			
			list->MakeEmpty();
			for(i=0; i<nitems; i++)
				delete items[i];			
*/						
		}
		
		void __cdecl
			Java_java_awt_Toolkit_lstRemoveItem ( JNIEnv* env, jclass clazz, KList* wnd, jint idx)
		{
UNIMPLEMENTED((
			SendMessage( wnd->hwnd, LB_DELETESTRING, (WPARAM)idx, 0);
			wnd->updateHScroll();
));
			if (wnd->lock()) {
				BListItem * item = wnd->NATIVE()->RemoveItem(idx);
				delete item;  // !!! 
				wnd->unlock();
			}
			wnd->updateHScroll();
		}
		
		void __cdecl
			Java_java_awt_Toolkit_lstReplaceItem ( JNIEnv* env, jclass clazz, KList* wnd, jstring item, jint idx)
		{
/**
			TCHAR* buf = java2WinString( env, X, item);
			SendMessage( wnd->hwnd, LB_DELETESTRING, (WPARAM)idx, 0);
			SendMessage( wnd->hwnd, LB_INSERTSTRING, idx, (LPARAM)buf);
			wnd->updateHScroll();
**/
			char* buf = java2CString( env, X, item);
			DBG(AWT, ("lstReplaceItem %p %d : %s\n", wnd, idx, buf));
			if (wnd->lock()) {
				wnd->NATIVE()->ReplaceItem(idx, new BStringItem(buf));
				wnd->unlock();
			}
			wnd->updateHScroll();
		}
		
		void __cdecl
			Java_java_awt_Toolkit_lstSetMultipleMode ( JNIEnv* env, jclass clazz, KList* wnd, jboolean multi)
		{
/**
			int cs = GetWindowLong( wnd->hwnd, GWL_STYLE);
			
			wnd->multiSel = mode;
			if ( mode ) cs |= LBS_MULTIPLESEL;
			else		cs &= ~LBS_MULTIPLESEL;
			
			SetWindowLong( wnd->hwnd, GWL_STYLE, cs);
**/
			if (wnd->lock()) {
				wnd->setMode( multi);
				wnd->unlock();
			}
		}
		
		
}
