/**
 * @file menu.cpp 
 * Menu support functions
 *
 * Copyright (c) 1998-2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * @author T. Paoletti <tpaoletti@users.sourceforge.net>,
 *         <your name here>
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "toolkit.hpp"
#include "cmnwnd.hpp"


#define USHORT		uint16

extern BMessage * KWM_CONTEXTMENU;

const char * MenuBarName = "_root_mbar_";
#define EVENT_MENU	'KMnu'

/** 
    Implementation of native menus.
	This implementation is based on the Win32 menu model, 
    and is not particularly clean.
    
    @todo rewrite KMenu and Java counterpart ? 
*/
class KMenu {
public:	
	BMenu 		*fMenu;
	jboolean	isPopup;
	/** @todo do we need fParent ? */
	KWnd*		fParent;		
	
	KMenu( jboolean popup) {
		isPopup = popup;
		if ( isPopup ) {
			fMenu = new BMenu("KMenu,popup");
		}
		else {
			fMenu = new BMenu("KMenu");
		}
		DBG(AWT, ("new KMenu %p\n", fMenu));
	}
	
	bool lock() {
		return fMenu->LockLooperWithTimeout(5e6)==B_OK;
	}
	
	void unlock() {
		fMenu->UnlockLooper();
	}
	
	BMenu * NATIVE() {
		return fMenu;
	}
	
	void KMenu::removeItem( USHORT id) {
UNIMPLEMENTED((
		DeleteMenu( fMenu, id, MF_BYCOMMAND);
));
	}

	/** @todo rewrite so that it doesn't scan all menu items looking for id
	*/
	BMenuItem * 
	KMenu::findItem(jobject id ) {
			BMenuItem * mi = NULL;
			BMessage  * msg;
			int32 i;
			int32 n = NATIVE()->CountItems();
			for(i=0; i<n; i++) {
				mi = NATIVE()->ItemAt(i);
				if ((msg = mi->Message()) != NULL) {
					if (msg->FindInt32("Kid") == (int32)id) 
						return mi;
				} 
			}
			return NULL;
	}
	
	/* !!! TEST AGAIN */
	void KMenu::checkItem( jobject id, jboolean state) {
		if (lock()) {
			BMenuItem * mi = findItem(id);
			if (mi)
				mi->SetMarked(state);
			unlock();
		}
	}
	
	/* !!! TEST AGAIN */
	void KMenu::enableItem( jobject id, jboolean state) {
		if (lock()) {
			BMenuItem * mi = findItem(id);
			if (mi) 
				mi->SetEnabled(state);
			unlock();
		}
	}
	
	void KMenu::insertItem( char* item, jobject id, int pos, BMenu * sub) {
UNIMPLEMENTED((

#if !defined(UNDER_CE)
		MENUITEMINFO mi;
		mi.cbSize = sizeof( MENUITEMINFO);
		mi.fMask = MIIM_ID;
		mi.wID	 = (USHORT)id;
		
		if ( (item == NULL) || (*item == '-') ) {
			mi.fType = MFT_SEPARATOR;
			mi.fMask |= MIIM_TYPE;
		}
		else if ( sub) {
			mi.fType = MFT_STRING;
			mi.fMask |= MIIM_TYPE | MIIM_SUBMENU;
			mi.hSubMenu = sub;
			mi.dwTypeData = item;
			mi.cch = _tcslen( item);
		}
		else {
			mi.fMask |= MIIM_DATA | MIIM_TYPE;
			mi.fType = MFT_STRING;
			mi.dwItemData = id;
			mi.dwTypeData = item;
			mi.cch = _tcslen( item);
		}
		if ( pos == -1 ) {
			InsertMenuItem( fMenu, mi.wID, FALSE, &mi);
		}
		else {
			InsertMenuItem( fMenu, pos, TRUE, &mi);
		}
#else
		/* CE doesn't have InsertMenuItem yet, revert to old Append/InsertMenu
		 */
        UINT flags = (!item || (*item=='-')) ? MF_SEPARATOR : (sub ? MF_POPUP : MF_STRING);
		UINT mid = (UINT) (sub ? (UINT)sub : id);

		if ( pos == -1 ) {
			AppendMenu( fMenu, flags, mid, item);
		}
		else {
			InsertMenu( fMenu, pos, flags, mid, item);
		}
#endif
));
		BMenuItem * mi;
		
		if ( (item == NULL) || (*item == '-') ) {
			mi = new BSeparatorItem();
		} 
		else if ( sub) {
			UNIMPLEMENTED(( /* item is not set */ ));
			sub->SetName(item);
			mi = new BMenuItem(sub, NULL);	/* !!! submenus don't send notifications, right ? */
		} else {
			BMessage * msg= new BMessage(EVENT_MENU);
			msg->AddInt32("Kid", (int32)id);		/* insert id of menu into the notification message */
			mi = new BMenuItem(item, msg);
		}
		if ( pos == -1 ) {
			NATIVE()->AddItem( mi);
		}
		else {
			NATIVE()->AddItem( mi, pos);
		}
	}
	
	static BMenuBar * KMenu::findMenuBar( KWnd* parent) {
		if (parent == NULL) return NULL;
		
		BView * view = parent->getWindow()->FindView(MenuBarName);
		return (BMenuBar *)view;
	}
	
	static BMenuBar * KMenu::insertMenuBar( KWnd* parent, int x, int y) {
		BRect r(x,y,x,y);
		BMenuBar * mbar = new BMenuBar(r, MenuBarName);
		parent->getView()->AddChild(mbar);
		return mbar;
	}
	
	void KMenu::assign( KWnd* parent, int x, int y) {
UNIMPLEMENTED((		
		if ( ! isPopup) {
#if !defined(UNDER_CE)
			SetMenu( parent->hwnd, fMenu);
#else
			/* CE doesn't attach menubars to Frames, but to "CommandBars", filled
			 * with menus and buttons.
			 */
			
			CommandBar_InsertMenubarEx(parent->hbar, NULL, (TCHAR*)fMenu, 0);
			fParent = parent;
#endif
			parent->updateInsets();
		}
		else {
			parent->popup = fMenu;
			x -= parent->insets.left;
			y -= parent->insets.top;
			PostMessage( parent->hwnd, KWM_CONTEXTMENU, 
				(WPARAM)parent->hwnd, MAKELPARAM( x, y) );
		}
));
		if ( ! isPopup) {
			BMenuBar * mbar = KMenu::findMenuBar(parent);

			if (mbar==NULL) {
				mbar = KMenu::insertMenuBar(parent, x, y);	
			}
				
			for(int32 n=0; n<fMenu->CountItems(); n++) {
				mbar->AddItem(fMenu->SubmenuAt(n));
			}
			//mbar->AddItem(fMenu);
			DBG(AWT, ("KMenu::assign: menubar %p, parent %p\n", mbar, parent));
			
			fParent = parent;
			parent->updateInsets();
		}
		else {
UNIMPLEMENTED((
			parent->popup = fMenu;
			x -= parent->insets.left;
			y -= parent->insets.top;
			PostMessage( parent->hwnd, KWM_CONTEXTMENU, 
				(WPARAM)parent->hwnd, MAKELPARAM( x, y) );
));
		}
	}
	
	KMenu::~KMenu() {
UNIMPLEMENTED((
		DestroyMenu( fMenu);
));
	}
	
	
};

extern "C" {
	
	void* __cdecl
		Java_java_awt_Toolkit_menuCreateMenu ( JNIEnv* env, jclass clazz, jboolean popup)
	{
		return ( new KMenu( popup));
	}
	
	void __cdecl
		Java_java_awt_Toolkit_menuRedraw ( JNIEnv* env, jclass clazz, KWnd* parent)
	{
UNIMPLEMENTED((
#if defined(UNDER_CE)
		CommandBar_DrawMenuBar(parent->hbar, 0);
#endif
));
	}
	
	void __cdecl
		Java_java_awt_Toolkit_menuInsertItem ( JNIEnv* env, jclass clazz, KMenu* menu, KMenu* sub, jstring item, int pos, jobject id)
	{
		char* lbl = item ? java2CString( env, X, item) : NULL;
		DBG(AWT, ("menuInsertItem: %p %s %p\n", menu, lbl, id));
		menu->insertItem( lbl, id, pos, sub ? sub->fMenu : NULL);
	}
	
	void __cdecl
		Java_java_awt_Toolkit_menuRemoveItem ( JNIEnv* env, jclass clazz, KMenu* menu, jobject id)
	{
		DBG(AWT, ("menuRemoveItem: %p %p\n", menu, id));
		menu->removeItem( (USHORT)id);
	}
	
	void __cdecl
		Java_java_awt_Toolkit_menuDeleteMenu ( JNIEnv* env, jclass clazz, KMenu* menu)
	{
		DBG(AWT, ("menuDeleteMenu: %p\n", menu));
UNIMPLEMENTED((
		delete( menu) ;
));
	}
	
	void __cdecl
		Java_java_awt_Toolkit_menuCheckItem ( JNIEnv* env, jclass clazz, KMenu* menu, jobject id, jboolean state)
	{
		DBG(AWT, ("menuCheckItem: %p %p %d\n", menu, id, state));
		menu->checkItem( id, state);
	}
	
	void __cdecl
		Java_java_awt_Toolkit_menuEnableItem ( JNIEnv* env, jclass clazz, KMenu* menu, jobject id, jboolean state)
	{
		DBG(AWT, ("menuEnableItem: %p %p %d\n", menu, id, state));
		menu->enableItem( id, state);
	}
	
	void __cdecl
		Java_java_awt_Toolkit_menuAssign ( JNIEnv* env, jclass clazz, KWnd* parent, KMenu* menu, jint x, jint y)
	{
UNIMPLEMENTED((
		if ( ! menu) {
#if !defined(UNDER_CE)
			SetMenu( parent->hwnd, NULL);
#else
			CommandBar_InsertMenubarEx(parent->hbar, appInstance, NULL, 0);
#endif
			parent->updateInsets();
		}
		else {
#if defined(UNDER_CE)
			if ( (parent->hbar == NULL) && !menu->isPopup ) {
				parent->hbar = CommandBar_Create(appInstance, parent->hwnd, 0);
			}
#endif
			menu->assign( parent, x, y);
		}
));
		DBG(AWT, ("menuAssign to %p (menu %p) at %d,%d\n", parent, menu, x, y));
		if ( ! menu) {
			BMenuBar * mbar = KMenu::findMenuBar(parent);
			if (mbar == NULL)
				mbar = KMenu::insertMenuBar(parent, x, y);
									
			parent->updateInsets();
		} else {
			menu->assign( parent, x, y);
		}
	}
	
}
