/**
 * @file scroll.cpp 
 * Scrollbar support
 *
 * Copyright (c) 1998-2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * @author T. Paoletti <tpaoletti@users.sourceforge.net>,
 *         <your name here>
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "toolkit.hpp"
#include "cmnwnd.hpp"

#include <ScrollBar.h>

#define UNIT_INCREMENT		1
#define UNIT_DECREMENT		2
#define BLOCK_DECREMENT		3
#define BLOCK_INCREMENT     4
#define TRACK				5

/** 
	Native widget implementing scroll bars
*/
class KScroll:public KWnd {
public:		
	int min;
	int max;
	int pos;
	int page;
	
	KScroll( BView * parent, jboolean vert) {
		BRect r(0,0,0,0);
		BView * target = NULL;
		orientation orient;
		if (vert) {
			orient = B_VERTICAL;
			r.right = B_V_SCROLL_BAR_WIDTH;
			r.bottom = 100;
		} else {
			orient = B_HORIZONTAL;
			r.right = 100;
			r.bottom = B_H_SCROLL_BAR_HEIGHT;
		} 
		BView * v = new BScrollBar(r, "KScroll", target, 0, 100, orient);
		setView(v);				
		//initScroll();
		parent->AddChild(v);
	}
	
	BScrollBar * NATIVE() {
		return (BScrollBar *)getView();
	}
	
	void setBounds(int x, int y, int width, int height) {
		KWnd* parent = NULL; // !!! = (KWnd*)GetWindowLong( GetParent( hwnd), GWL_USERDATA );
	
		DBG(AWT_GRA, ("setBounds: %p %d,%d,%d,%d\n", this, x,y,x+width,y+height));
		
		if (lock()) {		
			if ( parent ) {
				x -= (int)parent->insets.left;
				y -= (int)parent->insets.top;
			}
				
			NATIVE()->MoveTo( x, y);

			if (NATIVE()->Orientation()==B_HORIZONTAL) 
				height = (int)B_H_SCROLL_BAR_HEIGHT;				
			else
				width = (int)B_V_SCROLL_BAR_WIDTH;
			
			NATIVE()->ResizeTo( width, height); 
			finishResize();		// finish resize of widget
			unlock();
		}
	}
	
	bool KScroll::OnScroll( int trackPos, int op, bool vert) {
UNIMPLEMENTED((
		int jop = -1;
		
		switch ( op) {
		case SB_PAGEDOWN:
			if ( pos < max-page+1) {
				jop = BLOCK_INCREMENT;
				pos = min( pos+page, max-page+1);
			}
			break;
		case SB_PAGEUP:
			if ( pos > min) {
				jop = BLOCK_DECREMENT;
				pos = max( pos-page, min);
			}
			break;
		case SB_LINEDOWN:
			if ( pos < max-page+1 ) {
				jop = UNIT_INCREMENT;
				pos = min( pos+1, max-page+1);
			}
			break;
		case SB_LINEUP:
			if ( pos > min ) {
				jop = UNIT_DECREMENT;
				pos = max( pos-1, min);
			}
			break;
		case SB_TOP:
			pos = min;
			break;
		case SB_BOTTOM:
			pos = max-page;
			break;
		case SB_THUMBTRACK:
			jop = TRACK;
			pos = trackPos;
			break;
		case SB_ENDSCROLL:
			break;
		case SB_THUMBPOSITION:
			break;
		}
		
		if ( jop > -1 ) {
			setPos( pos);
			JniEnv->CallStaticObjectMethod( AdjustmentEvent, getAdjustmentEvent,
				srcIdx, jop, pos );
		}
));		
		return true;
	}

	void KScroll::initScroll() {
		setValues( 0, 0, 100, 10);
	}

	void KScroll::setScroll() {
UNIMPLEMENTED((
		if ( page < max-min ) {
			SetScrollInfo( hwnd, SB_CTL, &si, true);
		}
		else { 		//range completely visible
			EnableScrollBar( hwnd, SB_CTL, ESB_DISABLE_BOTH );
		}
));
	}

	int KScroll::getPos() {
		int pos = 0;
		if (lock()) {
			pos = (int) NATIVE()->Value();
			unlock();
		}
		return pos;
	}
	
	void KScroll::setPos( int pos) {
		if (lock()) {
			NATIVE()->SetValue(pos);
			setScroll();
			unlock();
		}
	}
	
	void KScroll::setValues( int pos, int min, int max, int page) {
		if (lock()) {
			NATIVE()->SetRange(min,max);
			NATIVE()->SetValue(pos);
			NATIVE()->SetSteps(1, page);	// !!! smallStep is not handled
			setScroll();
			unlock();
		}
	}
	
	void KScroll::setRange( int min, int max) {
		if (lock()) {
			NATIVE()->SetRange(min,max);
			setScroll();
			unlock();
		}
	}
	
	void KScroll::setPage( int page) {
		if (lock()) {
			NATIVE()->SetSteps(1, page);	// !!! smallStep is not handled
			setScroll();
			unlock();
		}
	}

	};
	
	
	
	
	/*************************
	*	exported functions
	**************************/
	extern "C" {
		
		void* __cdecl
			Java_java_awt_Toolkit_scrollCreateScrollbar ( JNIEnv* env, jclass clazz, KWnd* parent, jboolean vert)
		{
			return ( new KScroll( parent->getView(), vert));
		}
		
		void __cdecl
			Java_java_awt_Toolkit_scrollSetValues ( JNIEnv* env, jclass clazz, KScroll* wnd, jint pos, jint min, jint max, jint page)
		{
			wnd->setValues( pos, min, max, page);
		}
		
		void __cdecl
			Java_java_awt_Toolkit_scrollSetPos ( JNIEnv* env, jclass clazz, KScroll* wnd, jint pos)
		{
			wnd->setPos( pos);
		}
		
		void __cdecl
			Java_java_awt_Toolkit_scrollSetRange ( JNIEnv* env, jclass clazz, KScroll* wnd, jint min, jint max)
		{
			wnd->setRange( min, max);
		}
		
		void __cdecl
			Java_java_awt_Toolkit_scrollSetPage ( JNIEnv* env, jclass clazz, KScroll* wnd, jint page)
		{
			wnd->setPage( page);
		}
		
		void __cdecl
			Java_java_awt_Toolkit_scrollSetOrientation ( JNIEnv* env, jclass clazz, KScroll* wnd, jboolean vert)
		{
UNIMPLEMENTED((
			int cs = GetWindowLong( wnd->hwnd, GWL_STYLE);
			
			if ( vert )	{
				cs |= SBS_VERT;
				cs &= ~SBS_HORZ;
			}
			else {
				cs |= SBS_HORZ;
				cs &= ~SBS_VERT;
			}
			
			SetWindowLong( wnd->hwnd, GWL_STYLE, cs);
));
		}
		
	}
