/** 
 * @file tlk.cpp
 * Toolkit support functions.
 *
 * Copyright (c) 1998-2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */


#define MAIN

#include <AppKit.h>
#include <SupportKit.h>

#include "toolkit.hpp"
#include "tlkprops.hpp"

			/** MIME signature for the Kaffe VM application */
const char * app_signature = "application/x-vnd.kaffevm";

/* !!!
void
registerClasses()
{

	WNDCLASS	wc;
	
	// generic toplevel class
	wc.style = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
	wc.lpfnWndProc = WndProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = 0;
	wc.hIcon = APPICON;
	wc.lpszMenuName = 0;
	wc.hCursor = (HCURSOR)0;
	wc.hbrBackground = (HBRUSH)0;
	wc.lpszClassName = TEXT("window");
	
	RegisterClass ( &wc);
	
	//generic widget class ( used for canvas and panel)
	wc.style = CS_OWNDC;
	wc.hIcon = (HICON)0;
	wc.lpszClassName = TEXT("widget");
	
	RegisterClass ( &wc);
}
**/

/** Run the application loop */
int32 dispatcherLoop (void * arg) {
/* 
#if defined(D_LOOP_MODE)
	jmethodID initToolkit;

	MSG msg;
	
	X->dspTid = GetCurrentThreadId();

	DBG(AWT, ("Toolkit.initToolkit\n"));
    initToolkit = JniEnv->GetStaticMethodID( Tlk, "initToolkit", "()V");
    JniEnv->CallStaticVoidMethod( Tlk, initToolkit);
	DBG(AWT, ("out of Toolkit.initToolkit\n"));

    while ( GetMessage( &msg, NULL, 0, 0) == TRUE) {
		TranslateMessage( &msg);
		DispatchMessage( &msg);
	}
*/

	if (be_app==NULL) {
		DBG(AWT, ("app event loop starting\n"));
		/* we're running standalone, 
		   link to the app_server and run our own event loop */
		new BApplication(app_signature);
		be_app->Run();
		DBG(AWT, ("end of app event loop\n"));
		delete be_app;
	} else {
		DBG(AWT, ("toolkit is running embedded in another app\n"));
	}
/* #endif */
}

/********************************************************************************
* exported functions
*/

extern "C" {
	
	jint __cdecl
		Java_java_awt_Toolkit_tlkProperties ( JNIEnv* env, jclass clazz )
	{
		
#ifdef D_LOOP_MODE
		return (TLK_NATIVE_DISPATCHER_LOOP /* | TLK_IS_DISPATCH_EXCLUSIVE */ );
#else
		return (TLK_IS_BLOCKING /* | TLK_IS_DISPATCH_EXCLUSIVE */ );
#endif
	}
	
	jboolean __cdecl
		Java_java_awt_Toolkit_tlkInit ( JNIEnv* env, jclass clazz, jstring name )
	{
		DBG(AWT, ("tlkInit\n"));
		 
		X->nBuf = 128;
		X->buf = (char*)AWT_MALLOC( X->nBuf * sizeof(char));
		X->lastNativeWnd = (BWindow *)-1;
		
		Tlk = clazz;
		JniEnv = env;

/* !!!		
		registerClasses();
		X->wakeUp = CreateWindow( TEXT("window"), TEXT(""), WS_OVERLAPPEDWINDOW, 0, 0, 0, 0,
			NULL, NULL, NULL, NULL);

		X->display = GetDC( X->wakeUp);
**/
#ifdef D_LOOP_MODE
			/* in NATIVE_DISPATCHER_LOOP mode, the thread calling tlkInit blocks on the loop */
		dispatcherLoop(NULL);	
#else
			/* otherwise we run the application event loop in a new thread */
		thread_id dspLoop = spawn_thread(dispatcherLoop, "dispatcher", B_NORMAL_PRIORITY, NULL);
		resume_thread(dspLoop);
#endif
				
		return JNI_TRUE;
	}
	
	void __cdecl
		Java_java_awt_Toolkit_tlkTerminate ( JNIEnv* env, jclass clazz )
	{
/* !!!
		DeleteDC( X->display);
		if ( X->wakeUp ) {
			DestroyWindow( X->wakeUp);
			X->wakeUp = 0;
		}
***/
		DBG(AWT, ("tlkTerminate\n"));
  
		/* !!! must check if we are stand-alone */
  		/* unlink from the app_server */
		if (be_app->Lock()) {
			be_app->Quit();  
		}
	}
	
	
	jstring __cdecl
		Java_java_awt_Toolkit_tlkVersion ( JNIEnv* env, jclass clazz )
	{
		return env->NewStringUTF( "BeOS Native");
	}
	
	jint __cdecl
		Java_java_awt_Toolkit_tlkGetResolution ( JNIEnv* env, jclass clazz )
	{
		return 72; // !!! JUST A GUESS
	}
	
	jint __cdecl
		Java_java_awt_Toolkit_tlkGetScreenHeight ( JNIEnv* env, jclass clazz )
	{
		DBG(AWT, ("tlkGetScreenHeight\n"));
		return jint( BScreen().Frame().bottom );
	}
	
	jint __cdecl
		Java_java_awt_Toolkit_tlkGetScreenWidth ( JNIEnv* env, jclass clazz )
	{
		DBG(AWT, ("tlkGetScreenWidth\n"));
		return jint( BScreen().Frame().right );
	}
	
	jint __cdecl
		Java_java_awt_Toolkit_tlkGetMenubarHeight ( JNIEnv* env, jclass clazz )
	{
		return 20;	// !!! find a function for this
	}
	
	jint __cdecl
		Java_java_awt_Toolkit_tlkGetScrollbarWidth ( JNIEnv* env, jclass clazz )
	{
		return (jint)B_V_SCROLL_BAR_WIDTH;
	}
	
	void __cdecl
		Java_java_awt_Toolkit_tlkSync ( JNIEnv* env, jclass clazz )
	{
	}
	
	void __cdecl
		Java_java_awt_Toolkit_tlkFlush ( JNIEnv* env, jclass clazz )
	{
UNIMPLEMENTED((
		GFLUSH();
));
	}
	
	
	void __cdecl
		Java_java_awt_Toolkit_tlkBeep ( JNIEnv* env, jclass clazz )
	{
DBG(AWT, ("tlkBeep\n"));
		beep();
	}
	
		/** -
		
		    never been to India, never jumped from an airplane, never saw a Kaffe display banner */
	void __cdecl
		Java_java_awt_Toolkit_tlkDisplayBanner ( JNIEnv* env, jclass clazz, jstring banner )
	{
	}
	
}
