/**
 * keysyms.h - 
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */


#ifndef __keysyms_h
#define __keysyms_h

#define KEY_F1	0x70                               

/*
 * negative codes indicate control/function keys which should produce a non-zero
 * keyChar value (AKA "control-keys"), i.e. which should generate KEY_TYPED
 * events
 */

signed char FKeyCode [256] = {
0,
0x24,               /* HOME            B_HOME           0x01         */
0,
0,
0x23,               /* END             B_END            0x04  EOL        */
0x9b,               /* INSERT          B_INSERT         0x05  Insert, insert here    */
0,
0,
-'\b',              /* BACK_SPACE      B_BACKSPACE		0x08  back space, back char  */
-'\t',              /* TAB             B_TAB            0x09                         */
-'\n',              /* ENTER           B_ENTER        	0x0A  Linefeed, LF           */
0x21,               /* PAGE_UP         B_PAGE_UP   		0x0B  Prior, previous        */
0x22,               /* PAGE_DOWN       B_PAGE_DOWN  	0x0C  Next                   */
-'\n',              /* ENTER           B_ENTER          0x0D  Return, enter          */
0,
0,
0,	/* 0x10 reserved for B_FUNCTION_KEY */
0,
0,
0, // 0x13,               /* PAUSE           XK_Pause           0x13  Pause, hold            */
0, // 0x91,               /* SCROLL_LOCK     XK_Scroll_Lock     0x14                         */
0,                  /*                 XK_Sys_Req         0x15                         */
0,
0,
0,
0,
0,			/* B_SUBSTITUTE ? */
-0x1b,              /* ESCAPE          B_ESCAPE          0x1B                          */
0x25,               /* LEFT            B_LEFT_ARROW      0x1C  Move left, left arrow   */
0x27,               /* RIGHT           B_RIGHT_ARROW     0x1D  Move right, right arrow */
0x26,               /* UP              B_UP_ARROW        0x1E  Move up, up arrow       */
0x28,               /* DOWN            B_DOWN_ARROW      0x1E  Move down, down arrow   */
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,  /* 0x20 */
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,  /* 0x30               0..9, A..Z have 3rd byte == 0  */
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,  /* 0x40                 -> overlaps (Z = 0x5a)       */
0, // 0x24,               /* HOME            XK_Home            0x50                         */
0, // 0x25,               /* LEFT            XK_Left            0x51  Move left, left arrow  */
0, // 0x26,               /* UP              XK_Up              0x52  Move up, up arrow      */
0, // 0x27,               /* RIGHT           XK_Right           0x53  Move right, right arrow */
0, // 0x28,               /* DOWN            XK_Down            0x54  Move down, down arrow  */
0, // 0x21,               /* PAGE_UP         XK_Prior Page_Up   0x55  Prior, previous        */
0, // 0x22,               /* PAGE_DOWN       XK_Next Page_Down  0x56  Next                   */
0, // 0x23,               /* END             XK_End             0x57  EOL                    */
0,                  /*                 XK_Begin           0x58  BOL                    */
0,
0,
0,
0,
0,
0,
0,
0,                  /*                 XK_Select          0x60  Select, mark           */
0, // 0x9a,               /* PRINTSCREEN     XK_Print           0x61                         */
0,                  /*                 XK_Execute         0x62  Execute, run, do       */
0, // 0x9b,               /* INSERT          XK_Insert          0x63  Insert, insert here    */
0,
0,                  /*                 XK_Undo            0x65  Undo, oops             */
0,                  /*                 XK_Redo            0x66  redo, again            */
0,                  /*                 XK_Menu            0x67                         */
0,                  /*                 XK_Find            0x68  Find, search           */
0, // 0x03,               /* CANCEL          XK_Cancel          0x69  Cancel, stop, abort, exit */
0, // 0x9c,               /* HELP            XK_Help            0x6A  Help                   */
0,                  /*                 XK_Break           0x6B                         */
0,
0,
0,
0,
0,                                                     /* 0x70 */
0,
0,
0,
0,
0,
0,
0,
0,
0,
0,
0,
0,
0,
0, // 0x1f,               /* MODECHANGE      XK_Mode_switch, script_switch    0x7E  Character set switch   */
-0x7f,               /* DELETE           B_DELETE          0x7F  Delete, rubout         */
0, // -' ',               /* SPACE           XK_KP_Space        0x80  space                  */
0,
0,
0,
0,
0,
0,
0,
0,
0, // -'\t',               /* TAB             XK_KP_Tab          0x89                         */
0,
0,
0,
0, // -'\n',               /* ENTER           XK_KP_Enter        0x8D  enter                  */
0,
0,
0,
0, // 0x70,              /* F1               XK_KP_F1           0x91  PF1, KP_A, ...         */
0, // 0x71,              /* F2               XK_KP_F2           0x92                         */
0, // 0x72,              /* F3               XK_KP_F3           0x93                         */
0, // 0x73,              /* F4               XK_KP_F4           0x94                         */
0, // 0x24,              /* HOME             XK_KP_Home         0x95                         */
0, // 0x25,              /* LEFT             XK_KP_Left         0x96                         */
0, // 0x26,              /* UP               XK_KP_Up           0x97                         */
0, // 0x27,              /* RIGHT            XK_KP_Right        0x98                         */
0, // 0x28,              /* DOWN             XK_KP_Down         0x99                         */
0, // 0x21,              /* PAGE_UP          XK_KP_Prior Page_Up 0x9A                        */
0, // 0x22,              /* PAGE_DOWN        XK_KP_Next Page_Down 0x9B                       */
0, // 0x23,              /* END              XK_KP_End          0x9C                         */
0,                 /*                  XK_KP_Begin        0x9D                         */
0, // 0x9b,              /* INSERT           XK_KP_Insert       0x9E                         */
0, // -0x7f,              /* DELETE           XK_KP_Delete       0x9F                         */
0,                                                     /* 0xA0 */
0,
0,
0,
0,
0,
0,
0,
0,
0,
0, // 0x6a,              /* MULTIPLY         XK_KP_Multiply     0xAA                         */
0, // 0x6b,              /* ADD              XK_KP_Add          0xAB                         */
0, // 0x6c,              /* SEPARATER        XK_KP_Separator    0xAC  separator, often comma */
0, // 0x6d,              /* SUBTRACT         XK_KP_Subtract     0xAD                         */
0, // 0x6e,              /* DECIMAL          XK_KP_Decimal      0xAE                         */
0, // 0x6f,              /* DIVIDE           XK_KP_Divide       0xAF                         */
0, // 0x60,              /* NUMPAD0          XK_KP_0            0xB0                         */
0, // 0x61,              /* NUMPAD1          XK_KP_1            0xB1                         */
0, // 0x62,              /* NUMPAD2          XK_KP_2            0xB2                         */
0, // 0x63,              /* NUMPAD3          XK_KP_3            0xB3                         */
0, // 0x64,              /* NUMPAD4          XK_KP_4            0xB4                         */
0, // 0x65,              /* NUMPAD5          XK_KP_5            0xB5                         */
0, // 0x66,              /* NUMPAD6          XK_KP_6            0xB6                         */
0, // 0x67,              /* NUMPAD7          XK_KP_7            0xB7                         */
0, // 0x68,              /* NUMPAD8          XK_KP_8            0xB8                         */
0, // 0x69,              /* NUMPAD9          XK_KP_9            0xB9                         */
0,
0,
0,
0, // 0x3d,              /* EQUALS           XK_KP_Equal        0xBD  equals                 */
0, // 0x70,              /* F1               XK_F1              0xBE                         */
0, // 0x71,              /* F2               XK_F2              0xBF                         */
0, // 0x72,              /* F3               XK_F3              0xC0                         */
0, // 0x73,              /* F4               XK_F4              0xC1                         */
0, // 0x74,              /* F5               XK_F5              0xC2                         */
0, // 0x75,              /* F6               XK_F6              0xC3                         */
0, // 0x76,              /* F7               XK_F7              0xC4                         */
0, // 0x77,              /* F8               XK_F8              0xC5                         */
0, // 0x78,              /* F9               XK_F9              0xC6                         */
0, // 0x79,              /* F10              XK_F10             0xC7                         */
0, // 0x7a,              /* F11              XK_F11 L1          0xC8                         */
0, // 0x7b,              /* F12              XK_F12             0xC9                         */
0,
0,
0,
0,
0,
0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,                       /* 0xD0 */
0,                 /*                                     0xE0 */
0, // 0x10,              /* SHIFT            XK_Shift_L         0xE1                        */
0, // 0x10,              /* SHIFT            XK_Shift_R         0xE2                        */
0, // 0x11,              /* CONTROL          XK_Control_L       0xE3                        */
0, // 0x11,              /* CONTROL          XK_Control_R       0xE4                        */
0, // 0x14,              /* CAPS_LOCK        XK_Caps_Lock       0xE5                        */
0,                 /*                  XK_Shift_Lock      0xE6                        */
0, // 0x9d,              /* META             XK_Meta_L          0xE7                        */
0, // 0x9d,              /* META             XK_Meta_R          0xE8                        */
0, // 0x12,              /* ALT              XK_Alt_L           0xE9                        */
0, // 0x12,              /* ALT              XK_Alt_R           0xEA                        */
0,                 /*                  XK_Super_L         0xEB                        */
0,                 /*                  XK_Super_R         0xEC                        */
0,                 /*                  XK_Hyper_L         0xED                        */
0,                 /*                  XK_Hyper_R         0xEE                        */
0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,                         /* 0xF0 */
0 // -0x7f               /* DELETE           XK_Delete          0xFF  Delete, rubout         */
};


#endif
